# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import tempfile
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import yaml

from anemoi.datasets.create import creator_factory


class TestingContext:
    pass


def create_dataset(
    *,
    config: Union[str, Dict[str, Any]],
    output: Optional[str],
    delta: Optional[List[str]] = None,
    is_test: bool = False,
) -> str:
    """Create a dataset based on the provided configuration.

    Parameters
    ----------
    config : Union[str, Dict[str, Any]]
        The configuration for the dataset. Can be a path to a YAML file or a dictionary.
    output : Optional[str]
        The output path for the dataset. If None, a temporary directory will be created.
    delta : Optional[List[str]], optional
        List of delta for secondary statistics, by default None.
    is_test : bool, optional
        Flag indicating if the dataset creation is for testing purposes, by default False.

    Returns
    -------
    str
        The path to the created dataset.
    """
    if isinstance(config, dict):
        temp_file = tempfile.NamedTemporaryFile(mode="w", suffix=".yaml")
        yaml.dump(config, temp_file)
        config = temp_file.name

    if output is None:
        output = tempfile.mkdtemp(suffix=".zarr")

    creator_factory("init", config=config, path=output, overwrite=True, test=is_test).run()
    creator_factory("load", path=output).run()
    creator_factory("finalise", path=output).run()
    creator_factory("patch", path=output).run()

    if delta is not None:
        creator_factory("init_additions", path=output, delta=delta).run()
        creator_factory("run_additions", path=output, delta=delta).run()
        creator_factory("finalise_additions", path=output, delta=delta).run()

    creator_factory("cleanup", path=output).run()

    if delta is not None:
        creator_factory("cleanup", path=output, delta=delta).run()

    creator_factory("verify", path=output).run()

    return output



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from . import MissingDateError
from .dataset import Dataset
from .dataset import FullIndex
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Merge(Combined):
    """A class to merge multiple datasets along the dates axis, handling gaps in dates if allowed."""

    def __init__(self, datasets: List[Dataset], allow_gaps_in_dates: bool = False) -> None:
        """Initialize the Merge object.

        Parameters
        ----------
        datasets : List[Dataset]
            List of datasets to merge.
        allow_gaps_in_dates : bool, optional
            Whether to allow gaps in dates. Defaults to False.
        """
        super().__init__(datasets)

        self.allow_gaps_in_dates = allow_gaps_in_dates

        dates = dict()  # date -> (dataset_index, date_index)

        for i, d in enumerate(datasets):
            for j, date in enumerate(d.dates):
                date = date.astype(object)
                if date in dates:

                    d1 = datasets[dates[date][0]]  # Selected
                    d2 = datasets[i]  # The new one

                    if j in d2.missing:
                        # LOG.warning(f"Duplicate date {date} found in datasets {d1} and {d2}, but {date} is missing in {d}, ignoring")
                        continue

                    k = dates[date][1]
                    if k in d1.missing:
                        # LOG.warning(f"Duplicate date {date} found in datasets {d1} and {d2}, but {date} is missing in {d}, ignoring")
                        dates[date] = (i, j)  # Replace the missing date with the new one
                        continue

                    raise ValueError(f"Duplicate date {date} found in datasets {d1} and {d2}")
                else:
                    dates[date] = (i, j)

        all_dates = sorted(dates)
        start = all_dates[0]
        end = all_dates[-1]

        frequency = min(d2 - d1 for d1, d2 in zip(all_dates[:-1], all_dates[1:]))

        date = start
        indices = []
        _dates = []

        self._missing_index = len(datasets)

        while date <= end:
            if date not in dates:
                if self.allow_gaps_in_dates:
                    dates[date] = (self._missing_index, -1)
                else:
                    raise ValueError(
                        f"merge: date {date} not covered by dataset. Start={start}, end={end}, frequency={frequency}"
                    )

            indices.append(dates[date])
            _dates.append(date)
            date += frequency

        self._dates = np.array(_dates, dtype="datetime64[s]")
        self._indices = np.array(indices)
        self._frequency = frequency.astype(object)

    def __len__(self) -> int:
        """Get the number of dates in the merged dataset.

        Returns
        -------
        int
            Number of dates.
        """
        return len(self._dates)

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Get the dates of the merged dataset."""
        return self._dates

    @property
    def frequency(self) -> datetime.timedelta:
        """Get the frequency of the dates in the merged dataset."""
        return self._frequency

    @cached_property
    def missing(self) -> Set[int]:
        """Get the indices of missing dates in the merged dataset."""
        # TODO: optimize
        result: Set[int] = set()

        for i, (dataset, row) in enumerate(self._indices):
            if dataset == self._missing_index:
                result.add(i)
                continue

            if row in self.datasets[dataset].missing:
                result.add(i)

        return result

    def check_same_lengths(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the lengths of two datasets are the same. (Disabled for merging).

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the dates of two datasets are the same. (Disabled for merging).

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        # Turned off because we are concatenating along the dates axis
        pass

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Check if two datasets are compatible for merging.

        Parameters
        ----------
        d1 : Dataset
            First dataset.
        d2 : Dataset
            Second dataset.
        """
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def tree(self) -> Node:
        """Get the tree representation of the merged dataset.

        Returns
        -------
        Node
            Tree representation of the merged dataset.
        """
        return Node(self, [d.tree() for d in self.datasets], allow_gaps_in_dates=self.allow_gaps_in_dates)

    def metadata_specific(self) -> Dict[str, Any]:
        """Get the specific metadata for the merged dataset.

        Returns
        -------
        Dict[str, Any]
            Specific metadata.
        """
        return {"allow_gaps_in_dates": self.allow_gaps_in_dates}

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get the item at the specified index.

        Parameters
        ----------
        n : FullIndex
            Index to retrieve.

        Returns
        -------
        NDArray[Any]
            Retrieved item.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        dataset, row = self._indices[n]

        if dataset == self._missing_index:
            raise MissingDateError(f"Date {self.dates[n]} is missing (index={n})")

        return self.datasets[dataset][int(row)]

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get the item at the specified tuple index.

        Parameters
        ----------
        index : TupleIndex
            Tuple index to retrieve.

        Returns
        -------
        NDArray[Any]
            Retrieved item.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 0, slice(None))
        result = self._get_slice(previous)
        return apply_index_to_slices_changes(result[index], changes)

    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Get the items in the specified slice.

        Parameters
        ----------
        s : slice
            Slice to retrieve.

        Returns
        -------
        NDArray[Any]
            Retrieved items.
        """
        return np.stack([self[i] for i in range(*s.indices(self._len))])


def merge_factory(args: Tuple, kwargs: Dict[str, Any]) -> Dataset:
    """Factory function to create a merged dataset.

    Parameters
    ----------
    args : Tuple
        Positional arguments.
    kwargs : Dict[str, Any]
        Keyword arguments.

    Returns
    -------
    Dataset
        Merged dataset.
    """
    datasets = kwargs.pop("merge")

    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    datasets = [_open(e) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    allow_gaps_in_dates = kwargs.pop("allow_gaps_in_dates", False)

    return Merge(datasets, allow_gaps_in_dates=allow_gaps_in_dates)._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import Dict
from typing import List

import earthkit.data as ekd
from earthkit.data.core.fieldlist import MultiFieldList
from earthkit.data.sources.url import download_and_cache

from .legacy import legacy_source
from .patterns import iterate_patterns
from .xarray import load_one


@legacy_source(__file__)
def execute(context: Any, dates: Any, record_id: str, file_key: str, *args: Any, **kwargs: Any) -> ekd.FieldList:
    """Executes the download and processing of files from Zenodo.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    dates : Any
        The dates for which the data is required.
    record_id : str
        The Zenodo record ID.
    file_key : str
        The key to identify the file.
    *args : Any
        Additional arguments.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    MultiFieldList
        A list of fields loaded from the downloaded files.
    """
    import requests

    result: List[Any] = []

    URLPATTERN = "https://zenodo.org/api/records/{record_id}"
    url = URLPATTERN.format(record_id=record_id)
    r = requests.get(url)
    r.raise_for_status()
    record: Dict[str, Any] = r.json()

    urls: Dict[str, str] = {}
    for file in record["files"]:
        urls[file["key"]] = file["links"]["self"]

    for url, dates in iterate_patterns(file_key, dates, **kwargs):
        if url not in urls:
            continue

        path = download_and_cache(urls[url])
        result.append(load_one("?", context, dates, path, options={}, flavour=None, **kwargs))

    return MultiFieldList(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import warnings
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from .action import Action
from .action import ActionContext
from .context import Context
from .result import Result
from .template import notify_result
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class StepResult(Result):
    """Represents the result of a step in the data processing pipeline."""

    def __init__(
        self, context: Context, action_path: List[str], group_of_dates: Any, action: Action, upstream_result: Result
    ) -> None:
        """Initialize a StepResult instance.

        Parameters
        ----------
        context
            The context in which the step is executed.
        action_path
            The path of actions leading to this step.
        group_of_dates
            The group of dates associated with this step.
        action
            The action associated with this step.
        upstream_result
            The result of the upstream step.
        """
        super().__init__(context, action_path, group_of_dates)
        assert isinstance(upstream_result, Result), type(upstream_result)
        self.upstream_result: Result = upstream_result
        self.action: Action = action

    @property
    @notify_result
    @trace_datasource
    def datasource(self) -> Any:
        """Retrieve the datasource associated with this step result."""
        raise NotImplementedError(f"Not implemented in {self.__class__.__name__}")


class StepAction(Action):
    """Represents an action that is part of a step in the data processing pipeline."""

    result_class: Optional[Type[StepResult]] = None

    def __init__(
        self, context: ActionContext, action_path: List[str], previous_step: Any, *args: Any, **kwargs: Any
    ) -> None:
        """Initialize a StepAction instance.

        Parameters
        ----------
        context
            The context in which the action is executed.
        action_path
            The path of actions leading to this step.
        previous_step
            The previous step in the pipeline.
        """
        super().__init__(context, action_path, *args, **kwargs)
        self.previous_step: Any = previous_step

    @trace_select
    def select(self, group_of_dates: Any) -> StepResult:
        """Select the result for a given group of dates.

        Parameters
        ----------
        group_of_dates
            The group of dates to select the result for.

        Returns
        -------
        unknown
            The result of the step.
        """
        return self.result_class(
            self.context,
            self.action_path,
            group_of_dates,
            self,
            self.previous_step.select(group_of_dates),
        )

    def __repr__(self) -> str:
        """Return a string representation of the StepAction instance.

        Returns
        -------
        unknown
            String representation of the instance.
        """
        return self._repr(self.previous_step, _inline_=str(self.kwargs))


def step_factory(config: Dict[str, Any], context: ActionContext, action_path: List[str], previous_step: Any) -> Any:
    """Factory function to create a step action based on the given configuration.

    Parameters
    ----------
    config
        The configuration dictionary for the step.
    context
        The context in which the step is executed.
    action_path
        The path of actions leading to this step.
    previous_step
        The previous step in the pipeline.

    Returns
    -------
    unknown
        An instance of a step action.
    """

    from .filter import FilterStepAction
    from .filter import FunctionStepAction

    assert isinstance(context, Context), (type, context)
    if not isinstance(config, dict):
        raise ValueError(f"Invalid input config {config}")

    config = deepcopy(config)
    assert len(config) == 1, config

    key = list(config.keys())[0]
    cls = dict(
        filter=FilterStepAction,
        # rename=RenameAction,
        # remapping=RemappingAction,
    ).get(key)

    if isinstance(config[key], list):
        args, kwargs = config[key], {}

    if isinstance(config[key], dict):
        args, kwargs = [], config[key]

    if isinstance(config[key], str):
        args, kwargs = [config[key]], {}

    if cls is not None:
        return cls(context, action_path, previous_step, *args, **kwargs)

    # Try filters from datasets filter registry
    from anemoi.transform.filters import filter_registry as transform_filter_registry

    from ..filters import create_filter as create_datasets_filter
    from ..filters import filter_registry as datasets_filter_registry

    if datasets_filter_registry.is_registered(key):

        if transform_filter_registry.is_registered(key):
            warnings.warn(f"Filter `{key}` is registered in both datasets and transform filter registries")

        filter = create_datasets_filter(None, config)
        return FunctionStepAction(context, action_path + [key], previous_step, key, filter)

    # Use filters from transform registry

    if transform_filter_registry.is_registered(key):
        from ..filters.transform import TransformFilter

        return FunctionStepAction(
            context, action_path + [key], previous_step, key, TransformFilter(context, key, config)
        )

    raise ValueError(f"Unknown step action `{key}`")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any
from typing import List

import earthkit.data as ekd

from .legacy import legacy_source
from .xarray import load_many


@legacy_source(__file__)
def execute(context: Any, dates: List[str], url: str, *args: Any, **kwargs: Any) -> ekd.FieldList:
    """Execute the data loading process.

    Parameters
    ----------
    context : Any
        The context in which the execution occurs.
    dates : List[str]
        List of dates for which data is to be loaded.
    url : str
        The URL from which data is to be loaded.
    *args : tuple
        Additional positional arguments.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    ekd.FieldList
        The loaded data.
    """
    return load_many("🇿", context, url, *args, **kwargs).execute(dates)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any
from typing import List

from earthkit.data import from_source

from .legacy import legacy_source


@legacy_source(__file__)
def forcings(context: Any, dates: List[str], template: str, param: str) -> Any:
    """Loads forcing data from a specified source.

    Parameters
    ----------
    context : object
        The context in which the function is executed.
    dates : list
        List of dates for which data is to be loaded.
    template : FieldList
        Template for the data source.
    param : str
        Parameter for the data source.

    Returns
    -------
    object
        Loaded forcing data.
    """
    context.trace("✅", f"from_source(forcings, {template}, {param}")
    return from_source("forcings", source_or_dataset=template, date=dates, param=param)


execute = forcings



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property
from typing import Any
from typing import Type

from earthkit.data import FieldList

from .function import FunctionContext
from .misc import _tidy
from .misc import assert_fieldlist
from .step import StepAction
from .step import StepResult
from .template import notify_result
from .trace import trace_datasource

LOG = logging.getLogger(__name__)


class FilterStepResult(StepResult):
    @property
    @notify_result
    @assert_fieldlist
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns the filtered datasource."""
        ds: FieldList = self.upstream_result.datasource
        ds = ds.sel(**self.action.kwargs)
        return _tidy(ds)


class FilterStepAction(StepAction):
    """Represents an action to filter a step result."""

    result_class: Type[FilterStepResult] = FilterStepResult


class StepFunctionResult(StepResult):
    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns the datasource after applying the function."""

        self.action.filter.context = FunctionContext(self)
        try:
            return _tidy(
                self.action.filter.execute(
                    self.upstream_result.datasource,
                    *self.action.args[1:],
                    **self.action.kwargs,
                )
            )

        except Exception:
            LOG.error(f"Error in {self.action.name}", exc_info=True)
            raise

    def _trace_datasource(self, *args: Any, **kwargs: Any) -> str:
        """Traces the datasource for the given arguments.

        Parameters
        ----------
        *args : Any
            The arguments.
        **kwargs : Any
            The keyword arguments.

        Returns
        -------
        str
            A string representation of the traced datasource.
        """
        return f"{self.action.name}({self.group_of_dates})"


class FunctionStepAction(StepAction):
    """Represents an action to apply a function to a step result."""

    result_class: Type[StepFunctionResult] = StepFunctionResult

    def __init__(
        self,
        context: object,
        action_path: list,
        previous_step: StepAction,
        name: str,
        filter: Any,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initializes a FunctionStepAction instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : list
            The action path.
        previous_step : StepAction
            The previous step action.
        *args : Any
            Additional arguments.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, action_path, previous_step, *args, **kwargs)
        self.name = name
        self.filter = filter



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import List

from earthkit.data.core.order import build_remapping

from ...dates.groups import GroupOfDates
from .context import Context
from .template import substitute

LOG = logging.getLogger(__name__)


class Action:
    """Represents an action to be performed within a given context.

    Attributes
    ----------
    context : ActionContext
        The context in which the action exists.
    kwargs : Dict[str, Any]
        Additional keyword arguments.
    args : Any
        Additional positional arguments.
    action_path : List[str]
        The action path.
    """

    def __init__(
        self, context: "ActionContext", action_path: List[str], /, *args: Any, **kwargs: Dict[str, Any]
    ) -> None:
        """Initialize an Action instance.

        Parameters
        ----------
        context : ActionContext
            The context in which the action exists.
        action_path : List[str]
            The action path.
        args : Any
            Additional positional arguments.
        kwargs : Dict[str, Any]
            Additional keyword arguments.
        """
        if "args" in kwargs and "kwargs" in kwargs:
            """We have:
               args = []
               kwargs = {args: [...], kwargs: {...}}
            move the content of kwargs to args and kwargs.
            """
            assert len(kwargs) == 2, (args, kwargs)
            assert not args, (args, kwargs)
            args = kwargs.pop("args")
            kwargs = kwargs.pop("kwargs")

        assert isinstance(context, ActionContext), type(context)
        self.context = context
        self.kwargs = kwargs
        self.args = args
        self.action_path = action_path

    @classmethod
    def _short_str(cls, x: str) -> str:
        """Shorten the string representation if it exceeds 1000 characters.

        Parameters
        ----------
        x : str
            The string to shorten.

        Returns
        -------
        str
            The shortened string.
        """
        x = str(x)
        if len(x) < 1000:
            return x
        return x[:1000] + "..."

    def _repr(self, *args: Any, _indent_: str = "\n", _inline_: str = "", **kwargs: Any) -> str:
        """Generate a string representation of the Action instance.

        Parameters
        ----------
        args : Any
            Additional positional arguments.
        _indent_ : str, optional
            The indentation string, by default "\n".
        _inline_ : str, optional
            The inline string, by default "".
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        str
            The string representation.
        """
        more = ",".join([str(a)[:5000] for a in args])
        more += ",".join([f"{k}={v}"[:5000] for k, v in kwargs.items()])

        more = more[:5000]
        txt = f"{self.__class__.__name__}: {_inline_}{_indent_}{more}"
        if _indent_:
            txt = txt.replace("\n", "\n  ")
        return txt

    def __repr__(self) -> str:
        """Return the string representation of the Action instance.

        Returns
        -------
        str
            The string representation.
        """
        return self._repr()

    def select(self, dates: object, **kwargs: Any) -> None:
        """Select dates for the action.

        Parameters
        ----------
        dates : object
            The dates to select.
        kwargs : Any
            Additional keyword arguments.
        """
        self._raise_not_implemented()

    def _raise_not_implemented(self) -> None:
        """Raise a NotImplementedError indicating the method is not implemented."""
        raise NotImplementedError(f"Not implemented in {self.__class__.__name__}")

    def _trace_select(self, group_of_dates: GroupOfDates) -> str:
        """Trace the selection of a group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates to trace.

        Returns
        -------
        str
            The trace string.
        """
        return f"{self.__class__.__name__}({group_of_dates})"


class ActionContext(Context):
    """Represents the context in which an action is performed.

    Attributes
    ----------
    order_by : str
        The order by criteria.
    flatten_grid : bool
        Whether to flatten the grid.
    remapping : Dict[str, Any]
        The remapping configuration.
    use_grib_paramid : bool
        Whether to use GRIB parameter ID.
    """

    def __init__(self, /, order_by: str, flatten_grid: bool, remapping: Dict[str, Any], use_grib_paramid: bool) -> None:
        """Initialize an ActionContext instance.

        Parameters
        ----------
        order_by : str
            The order by criteria.
        flatten_grid : bool
            Whether to flatten the grid.
        remapping : Dict[str, Any]
            The remapping configuration.
        use_grib_paramid : bool
            Whether to use GRIB parameter ID.
        """
        super().__init__()
        self.order_by = order_by
        self.flatten_grid = flatten_grid
        self.remapping = build_remapping(remapping)
        self.use_grib_paramid = use_grib_paramid


def action_factory(config: Dict[str, Any], context: ActionContext, action_path: List[str]) -> Action:
    """Factory function to create an Action instance based on the configuration.

    Parameters
    ----------
    config : Dict[str, Any]
        The action configuration.
    context : ActionContext
        The context in which the action exists.
    action_path : List[str]
        The action path.

    Returns
    -------
    Action
        The created Action instance.
    """
    from .concat import ConcatAction
    from .data_sources import DataSourcesAction
    from .function import FunctionAction
    from .join import JoinAction
    from .pipe import PipeAction
    from .repeated_dates import RepeatedDatesAction

    # from .data_sources import DataSourcesAction

    assert isinstance(context, Context), (type, context)
    if not isinstance(config, dict):
        raise ValueError(f"Invalid input config {config}")
    if len(config) != 1:
        raise ValueError(f"Invalid input config. Expecting dict with only one key, got {list(config.keys())}")

    config = deepcopy(config)
    key = list(config.keys())[0]

    if isinstance(config[key], list):
        args, kwargs = config[key], {}
    elif isinstance(config[key], dict):
        args, kwargs = [], config[key]
    else:
        raise ValueError(f"Invalid input config {config[key]} ({type(config[key])}")

    cls = {
        "data_sources": DataSourcesAction,
        "data-sources": DataSourcesAction,
        "concat": ConcatAction,
        "join": JoinAction,
        "pipe": PipeAction,
        "function": FunctionAction,
        "repeated_dates": RepeatedDatesAction,
        "repeated-dates": RepeatedDatesAction,
    }.get(key)

    if cls is None:
        from ..sources import create_source

        source = create_source(None, substitute(context, config))
        return FunctionAction(context, action_path + [key], key, source)

    return cls(context, action_path + [key], *args, **kwargs)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

from .legacy import legacy_source


@legacy_source(__file__)
def execute(context, dates, params=None, **kwargs):
    import earthkit.data as ekd

    from anemoi.datasets import open_dataset

    ds = open_dataset(**kwargs)
    # dates_to_index = {date: i for i, date in enumerate(ds.dates)}

    indices = []
    for date in dates:
        idx = np.where(ds.dates == date)[0]
        if len(idx) == 0:
            continue
        indices.append((int(idx[0]), date))

    vars = ds.variables
    if params is None:
        params = vars

    if not isinstance(params, (list, tuple, set)):
        params = [params]

    params = set(params)
    results = []

    ensemble = ds.shape[2] > 1
    latitudes = ds.latitudes
    longitudes = ds.longitudes

    for idx, date in indices:

        metadata = dict(valid_datetime=date, latitudes=latitudes, longitudes=longitudes)

        for j, y in enumerate(ds[idx]):

            param = vars[j]
            if param not in params:
                continue

            # metadata['name'] = param
            # metadata['param_level'] = param
            metadata["param"] = param

            for k, e in enumerate(y):
                if ensemble:
                    metadata["number"] = k + 1

                metadata["values"] = e

                results.append(metadata.copy())

    print(results[0].keys())

    # "list-of-dicts" does support resolution
    results = ekd.from_source("list-of-dicts", results)

    # return new_fieldlist_from_list([new_field_from_latitudes_longitudes(x, latitudes, longitudes) for x in results])
    return results



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any
from typing import Dict
from typing import List

from earthkit.data import from_source

from .legacy import legacy_source


@legacy_source(__file__)
def constants(context: Any, dates: List[str], template: Dict[str, Any], param: str) -> Any:
    """Deprecated function to retrieve constants data.

    Parameters
    ----------
    context : Any
        The context object for tracing.
    dates : list of str
        List of dates for which data is required.
    template : dict of str to Any
        Template dictionary for the data source.
    param : str
        Parameter to retrieve.

    Returns
    -------
    Any
        Data retrieved from the source.
    """
    from warnings import warn

    warn(
        "The source `constants` is deprecated, use `forcings` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    context.trace("✅", f"from_source(constants, {template}, {param}")
    if len(template) == 0:
        raise ValueError("Forcings template is empty.")

    return from_source("forcings", source_or_dataset=template, date=dates, param=param)


execute: Any = constants



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import re
import warnings
from typing import Any
from typing import Callable
from typing import Optional
from typing import Union

import numpy as np
from anemoi.utils.dates import frequency_to_string
from numpy.typing import NDArray

LOG = logging.getLogger(__name__)


class DatasetName:
    """Class to validate and parse dataset names according to naming conventions."""

    def __init__(
        self,
        name: str,
        resolution: Optional[str] = None,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
        frequency: Optional[datetime.timedelta] = None,
    ):
        """Initialize a DatasetName instance.

        Parameters
        ----------
        name : str
            The name of the dataset.
        resolution : Optional[str], optional
            The resolution of the dataset.
        start_date : Optional[datetime.date], optional
            The start date of the dataset.
        end_date : Optional[datetime.date], optional
            The end date of the dataset.
        frequency : Optional[datetime.timedelta], optional
            The frequency of the dataset.
        """
        self.name = name
        self.parsed = self._parse(name)
        print("---------------")
        print(self.parsed)
        print("---------------")

        self.messages = []

        self.check_parsed()
        self.check_resolution(resolution)
        self.check_frequency(frequency)
        self.check_start_date(start_date)
        self.check_end_date(end_date)

        if self.messages:
            self.messages.append(f"{self} is parsed as :" + "/".join(f"{k}={v}" for k, v in self.parsed.items()))

    @property
    def error_message(self) -> str:
        """Generate an error message based on the collected messages."""
        out = " And ".join(self.messages)
        if out:
            out[0].upper() + out[1:]
        return out

    def raise_if_not_valid(self, print: Callable = print) -> None:
        """Raise a ValueError if the dataset name is not valid.

        Parameters
        ----------
        print : Callable
            The function to use for printing messages.
        """
        if self.messages:
            for m in self.messages:
                print(m)
            raise ValueError(self.error_message)

    def _parse(self, name: str) -> dict:
        """Parse the dataset name into its components.

        Parameters
        ----------
        name : str
            The name of the dataset.

        Returns
        -------
        dict
            The parsed components of the dataset name.
        """
        pattern = r"^(\w+)-([\w-]+)-(\w+)-(\w+)-(\d\d\d\d)-(\d\d\d\d)-(\d+h|\d+m)-v(\d+)-?([a-zA-Z0-9-]+)?$"
        match = re.match(pattern, name)

        if not match:
            raise ValueError(f"the dataset name '{name}' does not follow naming convention. Does not match {pattern}")

        parsed = {}
        if match:
            keys = [
                "purpose",
                "labelling",
                "source",
                "resolution",
                "start_date",
                "end_date",
                "frequency",
                "version",
                "additional",
            ]
            parsed = {k: v for k, v in zip(keys, match.groups())}

        return parsed

    def __str__(self) -> str:
        """Return the string representation of the dataset name."""
        return self.name

    def check_parsed(self) -> None:
        """Check if the dataset name was parsed correctly."""
        if not self.parsed:
            self.messages.append(
                f"the dataset name {self} does not follow naming convention. "
                "See here for details: "
                "https://anemoi-registry.readthedocs.io/en/latest/naming-conventions.html"
            )

    def check_resolution(self, resolution: Optional[str]) -> None:
        """Check if the resolution matches the expected format.

        Parameters
        ----------
        resolution : str or None
            The expected resolution.
        """
        if self.parsed.get("resolution") and self.parsed["resolution"][0] not in "0123456789on":
            self.messages.append(
                f"the resolution {self.parsed['resolution'] } should start "
                f"with a number or 'o' or 'n' in the dataset name {self}."
            )

        if resolution is None:
            return
        resolution_str = str(resolution).replace(".", "p").lower()
        self._check_missing("resolution", resolution_str)
        self._check_mismatch("resolution", resolution_str)

    def check_frequency(self, frequency: Optional[datetime.timedelta]) -> None:
        """Check if the frequency matches the expected format.

        Parameters
        ----------
        frequency : datetime.timedelta or None
            The expected frequency.
        """
        if frequency is None:
            return
        frequency_str = frequency_to_string(frequency)
        self._check_missing("frequency", frequency_str)
        self._check_mismatch("frequency", frequency_str)

    def check_start_date(self, start_date: Optional[datetime.date]) -> None:
        """Check if the start date matches the expected format.

        Parameters
        ----------
        start_date : datetime.date or None
            The expected start date.
        """
        if start_date is None:
            return
        start_date_str = str(start_date.year)
        self._check_missing("start_date", start_date_str)
        self._check_mismatch("start_date", start_date_str)

    def check_end_date(self, end_date: Optional[datetime.date]) -> None:
        """Check if the end date matches the expected format.

        Parameters
        ----------
        end_date : datetime.date or None
            The expected end date.
        """
        if end_date is None:
            return
        end_date_str = str(end_date.year)
        self._check_missing("end_date", end_date_str)
        self._check_mismatch("end_date", end_date_str)

    def _check_missing(self, key: str, value: str) -> None:
        """Check if a component is missing from the dataset name.

        Parameters
        ----------
        key : str
            The component key.
        value : str
            The expected value.
        """
        if value not in self.name:
            self.messages.append(f"the {key} is {value}, but is missing in {self.name}.")

    def _check_mismatch(self, key: str, value: str) -> None:
        """Check if a component value mismatches the expected value.

        Parameters
        ----------
        key : str
            The component key.
        value : str
            The expected value.
        """
        if self.parsed.get(key) and self.parsed[key] != value:
            self.messages.append(f"the {key} is {value}, but is {self.parsed[key]} in {self.name}.")


class StatisticsValueError(ValueError):
    """Custom error for statistics value issues."""

    pass


def check_data_values(
    arr: NDArray[Any], *, name: str, log: list = [], allow_nans: Union[bool, list, set, tuple, dict] = False
) -> None:
    """Check the values in the data array for validity.

    Parameters
    ----------
    arr : NDArray[Any]
        The data array to check.
    name : str
        The name of the data array.
    log : list, optional
        A list to log messages.
    allow_nans : bool or list or set or tuple or dict, optional
        Whether to allow NaNs in the data array.
    """
    shape = arr.shape

    if (isinstance(allow_nans, (set, list, tuple, dict)) and name in allow_nans) or allow_nans:
        arr = arr[~np.isnan(arr)]

    if arr.size == 0:
        warnings.warn(f"Empty array for {name} ({shape})")
        return

    assert arr.size > 0, (name, *log)

    min, max = arr.min(), arr.max()
    assert not (np.isnan(arr).any()), (name, min, max, *log)

    if min == 9999.0:
        warnings.warn(f"Min value 9999 for {name}")

    if max == 9999.0:
        warnings.warn(f"Max value 9999 for {name}")

    in_minus_1_plus_1 = dict(minimum=-1, maximum=1)
    limits = {
        "cos_latitude": in_minus_1_plus_1,
        "sin_latitude": in_minus_1_plus_1,
        "cos_longitude": in_minus_1_plus_1,
        "sin_longitude": in_minus_1_plus_1,
    }

    if name in limits:
        if min < limits[name]["minimum"]:
            warnings.warn(
                f"For {name}: minimum value in the data is {min}. "
                "Not in acceptable range [{limits[name]['minimum']} ; {limits[name]['maximum']}]"
            )
        if max > limits[name]["maximum"]:
            warnings.warn(
                f"For {name}: maximum value in the data is {max}. "
                "Not in acceptable range [{limits[name]['minimum']} ; {limits[name]['maximum']}]"
            )


def check_stats(minimum: float, maximum: float, mean: float, msg: str, **kwargs: Any) -> None:
    """Check if the mean value is within the min/max interval.

    Parameters
    ----------
    minimum : float
        The minimum value.
    maximum : float
        The maximum value.
    mean : float
        The mean value.
    msg : str
        The message to include in the error.
    **kwargs : Any
        Additional keyword arguments.
    """
    tolerance = (abs(minimum) + abs(maximum)) * 0.01
    if (mean - minimum < -tolerance) or (mean - minimum < -tolerance):
        raise StatisticsValueError(
            f"Mean is not in min/max interval{msg} : we should have {minimum} <= {mean} <= {maximum}"
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from scipy.spatial import cKDTree

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Combined
from .forwards import GivenAxis
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Concat(Combined):
    """A class to represent concatenated datasets."""

    def __len__(self) -> int:
        """Returns the total length of the concatenated datasets.

        Returns
        -------
        int
            Total length of the concatenated datasets.
        """
        return sum(len(i) for i in self.datasets)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Retrieves a tuple of data from the concatenated datasets based on the given index.

        Parameters
        ----------
        index : TupleIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Concatenated data array from the specified index.
        """
        index, changes = index_to_slices(index, self.shape)
        # print(index, changes)
        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(index[0], lengths)
        # print("slies", slices)
        result = [d[update_tuple(index, 0, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=0)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieves data from the concatenated datasets based on the given index.

        Parameters
        ----------
        n : FullIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data array from the concatenated datasets based on the index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        # TODO: optimize
        k = 0
        while n >= self.datasets[k]._len:
            n -= self.datasets[k]._len
            k += 1
        return self.datasets[k][n]

    @debug_indexing
    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Retrieves a slice of data from the concatenated datasets.

        Parameters
        ----------
        s : slice
            Slice object specifying the range of data to retrieve.

        Returns
        -------
        NDArray[Any]
            Concatenated data array from the specified slice.
        """
        result = []

        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(s, lengths)

        result = [d[i] for (d, i) in zip(self.datasets, slices) if i is not None]

        return np.concatenate(result)

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Check the compatibility of two datasets for concatenation.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def check_same_lengths(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the lengths of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the dates of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are concatenating along the dates axis
        pass

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Returns the concatenated dates of all datasets."""
        return np.concatenate([d.dates for d in self.datasets])

    @property
    def shape(self) -> Shape:
        """Returns the shape of the concatenated datasets."""
        return (len(self),) + self.datasets[0].shape[1:]

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the concatenated datasets.

        Returns
        -------
        Node
            A Node object representing the concatenated datasets.
        """
        return Node(self, [d.tree() for d in self.datasets])


class GridsBase(GivenAxis):
    """A base class for handling grids in datasets."""

    def __init__(self, datasets: List[Any], axis: int) -> None:
        """Initializes a GridsBase object.

        Parameters
        ----------
        datasets : List[Any]
            List of datasets.
        axis : int
            Axis along which to combine the datasets.
        """
        super().__init__(datasets, axis)
        # Shape: (dates, variables, ensemble, 1d-values)
        assert len(datasets[0].shape) == 4, "Grids must be 1D for now"

    def check_same_grid(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the grids of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # We don't check the grid, because we want to be able to combine
        pass

    def check_same_resolution(self, d1: Dataset, d2: Dataset) -> None:
        """Check if the resolutions of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # We don't check the resolution, because we want to be able to combine
        pass

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns metadata specific to the GridsBase object.

        Parameters
        ----------
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the GridsBase object.
        """
        return super().metadata_specific(
            multi_grids=True,
        )

    def collect_input_sources(self, collected: List[Any]) -> None:
        """Collects input sources from the datasets.

        Parameters
        ----------
        collected : List[Any]
            List to which the input sources are appended.
        """
        # We assume that,because they have different grids, they have different input sources
        for d in self.datasets:
            collected.append(d)
            d.collect_input_sources(collected)


class Grids(GridsBase):
    """A class to represent combined grids from multiple datasets."""

    # TODO: select the statistics of the most global grid?
    @property
    def latitudes(self) -> NDArray[Any]:
        """Returns the concatenated latitudes of all datasets."""
        return np.concatenate([d.latitudes for d in self.datasets])

    @property
    def longitudes(self) -> NDArray[Any]:
        """Returns the concatenated longitudes of all datasets."""
        return np.concatenate([d.longitudes for d in self.datasets])

    @property
    def grids(self) -> Tuple[Any, ...]:
        """Returns the grids of all datasets."""
        result = []
        for d in self.datasets:
            result.extend(d.grids)
        return tuple(result)

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the Grids object.

        Returns
        -------
        Node
            A Node object representing the Grids object.
        """
        return Node(self, [d.tree() for d in self.datasets], mode="concat")

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns:
        Dict[str, Any]: The metadata specific to the forwards subclass.
        """
        return {}


class Cutout(GridsBase):
    """A class to handle hierarchical management of Limited Area Models (LAMs) and a global dataset."""

    def __init__(
        self,
        datasets: List[Any],
        axis: int = 3,
        cropping_distance: float = 2.0,
        neighbours: int = 5,
        min_distance_km: Optional[float] = None,
        plot: Optional[bool] = None,
    ) -> None:
        """Initializes a Cutout object for hierarchical management of Limited Area
        Models (LAMs) and a global dataset, handling overlapping regions.

        Parameters
        ----------
        datasets : list
            List of LAM and global datasets.
        axis : int
            Concatenation axis, must be set to 3.
        cropping_distance : float
            Distance threshold in degrees for cropping cutouts.
        neighbours : int
            Number of neighboring points to consider when constructing masks.
        min_distance_km : float, optional
            Minimum distance threshold in km between grid points.
        plot : bool, optional
            Flag to enable or disable visualization plots.
        """
        super().__init__(datasets, axis)
        assert len(datasets) >= 2, "CutoutGrids requires at least two datasets"
        assert axis == 3, "CutoutGrids requires axis=3"
        assert cropping_distance >= 0, "cropping_distance must be a non-negative number"
        if min_distance_km is not None:
            assert min_distance_km >= 0, "min_distance_km must be a non-negative number"

        self.lams = datasets[:-1]  # Assume the last dataset is the global one
        self.globe = datasets[-1]
        self.axis = axis
        self.cropping_distance = cropping_distance
        self.neighbours = neighbours
        self.min_distance_km = min_distance_km
        self.plot = plot
        self.masks = []  # To store the masks for each LAM dataset
        self.global_mask = np.ones(self.globe.shape[-1], dtype=bool)

        # Initialize cumulative masks
        self._initialize_masks()

    def _initialize_masks(self) -> None:
        """Generate hierarchical masks for each LAM dataset by excluding overlapping regions with previous LAMs and creating a global mask for the global dataset.

        Raises
        ------
        ValueError
            If the global mask dimension does not match the global dataset grid points.
        """
        from anemoi.datasets.grids import cutout_mask

        for i, lam in enumerate(self.lams):
            assert len(lam.shape) == len(
                self.globe.shape
            ), "LAMs and global dataset must have the same number of dimensions"
            lam_lats = lam.latitudes
            lam_lons = lam.longitudes
            # Create a mask for the global dataset excluding all LAM points
            global_overlap_mask = cutout_mask(
                lam.latitudes,
                lam.longitudes,
                self.globe.latitudes,
                self.globe.longitudes,
                plot=False,
                min_distance_km=self.min_distance_km,
                cropping_distance=self.cropping_distance,
                neighbours=self.neighbours,
            )

            # Ensure the mask dimensions match the global grid points
            if global_overlap_mask.shape[0] != self.globe.shape[-1]:
                raise ValueError("Global mask dimension does not match global dataset grid " "points.")
            self.global_mask[~global_overlap_mask] = False

            # Create a mask for the LAM datasets hierarchically, excluding
            # points from previous LAMs
            lam_current_mask = np.ones(lam.shape[-1], dtype=bool)
            if i > 0:
                for j in range(i):
                    prev_lam = self.lams[j]
                    prev_lam_lats = prev_lam.latitudes
                    prev_lam_lons = prev_lam.longitudes
                    # Check for overlap by computing distances
                    if self.has_overlap(prev_lam_lats, prev_lam_lons, lam_lats, lam_lons):
                        lam_overlap_mask = cutout_mask(
                            prev_lam_lats,
                            prev_lam_lons,
                            lam_lats,
                            lam_lons,
                            plot=False,
                            min_distance_km=self.min_distance_km,
                            cropping_distance=self.cropping_distance,
                            neighbours=self.neighbours,
                        )
                        lam_current_mask[~lam_overlap_mask] = False
            self.masks.append(lam_current_mask)

    def has_overlap(
        self,
        lats1: NDArray[Any],
        lons1: NDArray[Any],
        lats2: NDArray[Any],
        lons2: NDArray[Any],
        distance_threshold: float = 1.0,
    ) -> bool:
        """Check for overlapping points between two sets of latitudes and longitudes within a specified distance threshold.

        Parameters
        ----------
        lats1 : NDArray[Any]
            Latitude array for the first dataset.
        lons1 : NDArray[Any]
            Longitude array for the first dataset.
        lats2 : NDArray[Any]
            Latitude array for the second dataset.
        lons2 : NDArray[Any]
            Longitude array for the second dataset.
        distance_threshold : float
            Distance in degrees to consider as overlapping.

        Returns
        -------
        bool
            True if any points overlap within the distance threshold, otherwise False.
        """
        # Create KDTree for the first set of points
        tree = cKDTree(np.vstack((lats1, lons1)).T)

        # Query the second set of points against the first tree
        distances, _ = tree.query(np.vstack((lats2, lons2)).T, k=1)

        # Check if any distance is less than the specified threshold
        return np.any(distances < distance_threshold)

    def __getitem__(self, index: FullIndex) -> NDArray[Any]:
        """Retrieve data from the masked LAMs and global dataset based on the given index.

        Parameters
        ----------
        index : FullIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data array from the masked datasets based on the index.
        """
        if isinstance(index, (int, slice)):
            index = (index, slice(None), slice(None), slice(None))
        return self._get_tuple(index)

    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Helper method that applies masks and retrieves data from each dataset according to the specified index.

        Parameters
        ----------
        index : TupleIndex
            Index specifying slices to retrieve data.

        Returns
        -------
        NDArray[Any]
            Concatenated data array from all datasets based on the index.
        """
        index, changes = index_to_slices(index, self.shape)
        # Select data from each LAM
        lam_data = [lam[index[:3]] for lam in self.lams]

        # First apply spatial indexing on `self.globe` and then apply the mask
        globe_data_sliced = self.globe[index[:3]]
        globe_data = globe_data_sliced[..., self.global_mask]

        # Concatenate LAM data with global data, apply the grid slicing
        result = np.concatenate(lam_data + [globe_data], axis=self.axis)[..., index[3]]

        return apply_index_to_slices_changes(result, changes)

    def collect_supporting_arrays(self, collected: List[Any], *path: Any) -> None:
        """Collect supporting arrays, including masks for each LAM and the global dataset.

        Parameters
        ----------
        collected : List[Any]
            List to which the supporting arrays are appended.
        *path : Any
            Variable length argument list specifying the paths for the masks.
        """
        # Append masks for each LAM
        for i, (lam, mask) in enumerate(zip(self.lams, self.masks)):
            collected.append((path + (f"lam_{i}",), "cutout_mask", mask))

        # Append the global mask
        collected.append((path + ("global",), "cutout_mask", self.global_mask))

    @cached_property
    def shape(self) -> Shape:
        """Returns the shape of the Cutout, accounting for retained grid points
        across all LAMs and the global dataset.
        """
        shapes = [np.sum(mask) for mask in self.masks]
        global_shape = np.sum(self.global_mask)
        total_shape = sum(shapes) + global_shape
        return tuple(self.lams[0].shape[:-1] + (int(total_shape),))

    def check_same_resolution(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the resolutions of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are combining different resolutions
        pass

    @property
    def grids(self) -> TupleIndex:
        """Returns the number of grid points for each LAM and the global dataset
        after applying masks.
        """
        grids = [np.sum(mask) for mask in self.masks]
        grids.append(np.sum(self.global_mask))
        return tuple(grids)

    @property
    def latitudes(self) -> NDArray[Any]:
        """Returns the concatenated latitudes of each LAM and the global dataset
        after applying masks.
        """
        lam_latitudes = np.concatenate([lam.latitudes[mask] for lam, mask in zip(self.lams, self.masks)])

        assert (
            len(lam_latitudes) + len(self.globe.latitudes[self.global_mask]) == self.shape[-1]
        ), "Mismatch in number of latitudes"

        latitudes = np.concatenate([lam_latitudes, self.globe.latitudes[self.global_mask]])
        return latitudes

    @property
    def longitudes(self) -> NDArray[Any]:
        """Returns the concatenated longitudes of each LAM and the global dataset
        after applying masks.
        """
        lam_longitudes = np.concatenate([lam.longitudes[mask] for lam, mask in zip(self.lams, self.masks)])

        assert (
            len(lam_longitudes) + len(self.globe.longitudes[self.global_mask]) == self.shape[-1]
        ), "Mismatch in number of longitudes"

        longitudes = np.concatenate([lam_longitudes, self.globe.longitudes[self.global_mask]])
        return longitudes

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the `Cutout` instance and
        its associated datasets.

        Returns
        -------
        Node
            A `Node` object representing the `Cutout` instance as the root
            node, with each dataset in `self.datasets` represented as a child
            node.
        """
        return Node(self, [d.tree() for d in self.datasets])

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Returns metadata specific to the Cutout object.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the Cutout object.
        """
        return {}


def grids_factory(args: Tuple[Any, ...], kwargs: dict) -> Dataset:
    """Factory function to create a Grids object.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    Dataset
        A Grids object.
    """
    if "ensemble" in kwargs:
        raise NotImplementedError("Cannot use both 'ensemble' and 'grids'")

    grids = kwargs.pop("grids")
    axis = kwargs.pop("axis", 3)

    assert len(args) == 0
    assert isinstance(grids, (list, tuple))

    datasets = [_open(e) for e in grids]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Grids(datasets, axis=axis)._subset(**kwargs)


def cutout_factory(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Dataset:
    """Factory function to create a Cutout object.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : Dict[str, Any]
        Keyword arguments.

    Returns
    -------
    Dataset
        A Cutout object.
    """
    if "ensemble" in kwargs:
        raise NotImplementedError("Cannot use both 'ensemble' and 'cutout'")

    cutout = kwargs.pop("cutout")
    axis = kwargs.pop("axis", 3)
    plot = kwargs.pop("plot", None)
    min_distance_km = kwargs.pop("min_distance_km", None)
    cropping_distance = kwargs.pop("cropping_distance", 2.0)
    neighbours = kwargs.pop("neighbours", 5)

    assert len(args) == 0
    assert isinstance(cutout, (list, tuple)), "cutout must be a list or tuple"

    datasets = [_open(e) for e in cutout]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Cutout(
        datasets,
        axis=axis,
        neighbours=neighbours,
        min_distance_km=min_distance_km,
        cropping_distance=cropping_distance,
        plot=plot,
    )._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


def make_rescale(
    variable: str, rescale: Union[Tuple[float, float], List[str], Dict[str, float]]
) -> Tuple[float, float]:
    """Create rescale parameters (scale and offset) based on the input rescale specification.

    Parameters
    ----------
    variable : str
        The variable name.
    rescale : Union[Tuple[float, float], List[str], Dict[str, float]]
        The rescale specification.

    Returns
    -------
    Tuple[float, float]
        The scale and offset values.
    """
    if isinstance(rescale, (tuple, list)):

        assert len(rescale) == 2, rescale

        if isinstance(rescale[0], (int, float)):
            return rescale

        from cfunits import Units

        u0 = Units(rescale[0])
        u1 = Units(rescale[1])

        x1, x2 = 0.0, 1.0
        y1, y2 = Units.conform([x1, x2], u0, u1)

        a = (y2 - y1) / (x2 - x1)
        b = y1 - a * x1

        return a, b

        return rescale

    if isinstance(rescale, dict):
        assert "scale" in rescale, rescale
        assert "offset" in rescale, rescale
        return rescale["scale"], rescale["offset"]

    assert False


class Rescale(Forwards):
    """A class to apply rescaling to dataset variables."""

    def __init__(
        self, dataset: Dataset, rescale: Dict[str, Union[Tuple[float, float], List[str], Dict[str, float]]]
    ) -> None:
        """Initialize the Rescale object.

        Parameters
        ----------
        dataset : Dataset
            The dataset to be rescaled.
        rescale : Dict[str, Union[Tuple[float, float], List[str], Dict[str, float]]]
            The rescale specifications.
        """
        super().__init__(dataset)
        for n in rescale:
            assert n in dataset.variables, n

        variables = dataset.variables

        self._a = np.ones(len(variables))
        self._b = np.zeros(len(variables))

        self.rescale = {}
        for i, v in enumerate(variables):
            if v in rescale:
                a, b = make_rescale(v, rescale[v])
                self.rescale[v] = a, b
                self._a[i], self._b[i] = a, b

        self._a = self._a[np.newaxis, :, np.newaxis, np.newaxis]
        self._b = self._b[np.newaxis, :, np.newaxis, np.newaxis]

        self._a = self._a.astype(self.forward.dtype)
        self._b = self._b.astype(self.forward.dtype)

    def tree(self) -> Node:
        """Get the tree representation of the rescale operation.

        Returns
        -------
        Node
            The tree representation.
        """
        return Node(self, [self.forward.tree()], rescale=self.rescale)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the rescale subclass.

        Returns
        -------
        Dict[str, Any]
            The metadata dictionary.
        """
        return dict(rescale=self.rescale)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Get a tuple of rescaled data based on the provided index.

        Parameters
        ----------
        index : TupleIndex
            The index to retrieve data.

        Returns
        -------
        NDArray[Any]
            The rescaled data.
        """
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, 1, slice(None))
        result = self.forward[index]
        result = result * self._a + self._b
        result = result[:, previous]
        result = apply_index_to_slices_changes(result, changes)
        return result

    @debug_indexing
    def __get_slice_(self, n: slice) -> NDArray[Any]:
        """Get a slice of rescaled data.

        Parameters
        ----------
        n : slice
            The slice to retrieve data.

        Returns
        -------
        NDArray[Any]
            The rescaled data.
        """
        data = self.forward[n]
        return data * self._a + self._b

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get an item or slice of rescaled data based on the provided index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve data.

        Returns
        -------
        NDArray[Any]
            The rescaled data.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self.__get_slice_(n)

        data = self.forward[n]

        return data * self._a[0] + self._b[0]

    @cached_property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Get the statistics of the rescaled data."""
        result = {}
        a = self._a.squeeze()
        assert np.all(a >= 0)

        b = self._b.squeeze()
        for k, v in self.forward.statistics.items():
            if k in ("maximum", "minimum", "mean"):
                result[k] = v * a + b
                continue

            if k in ("stdev",):
                result[k] = v * a
                continue

            raise NotImplementedError("rescale statistics", k)

        return result

    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Get the tendencies of the statistics of the rescaled data.

        Parameters
        ----------
        delta : Optional[datetime.timedelta]
            The time delta for tendencies calculation.

        Returns
        -------
        Dict[str, NDArray[Any]]
            The tendencies statistics dictionary.
        """
        result = {}
        a = self._a.squeeze()
        assert np.all(a >= 0)

        for k, v in self.forward.statistics_tendencies(delta).items():
            if k in ("maximum", "minimum", "mean", "stdev"):
                result[k] = v * a
                continue

            raise NotImplementedError("rescale tendencies statistics", k)

        return result



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import itertools
from abc import ABC
from abc import abstractmethod
from functools import cached_property
from typing import Any
from typing import Callable
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

from anemoi.datasets.dates import DatesProvider
from anemoi.datasets.dates import as_datetime


def _shorten(dates: Union[List[datetime.datetime], Tuple[datetime.datetime, ...]]) -> Union[str, List[str]]:
    """Shorten the list of dates for display.

    Args:
        dates (Union[List[datetime.datetime], Tuple[datetime.datetime, ...]]): The list of dates.

    Returns:
        Union[str, List[str]]: The shortened list of dates.
    """
    if isinstance(dates, (list, tuple)):
        dates = [d.isoformat() for d in dates]
        if len(dates) > 5:
            return f"{dates[0]}...{dates[-1]}"
    return dates


class GroupOfDates:
    """A class to represent a group of dates."""

    def __init__(self, dates: List[datetime.datetime], provider: DatesProvider, partial_ok: bool = False) -> None:
        assert isinstance(provider, DatesProvider), type(provider)
        assert isinstance(dates, list)

        self.dates = [as_datetime(_) for _ in dates]
        self.provider = provider
        self.partial_ok = partial_ok

    def __len__(self) -> int:
        """Return the number of dates in the group.

        Returns:
            int: The number of dates.
        """
        return len(self.dates)

    def __iter__(self) -> Iterator[datetime.datetime]:
        """Return an iterator over the dates in the group.

        Returns:
            Iterator[datetime.datetime]: The iterator over the dates.
        """
        return iter(self.dates)

    def __repr__(self) -> str:
        """Return a string representation of the group of dates.

        Returns:
            str: The string representation.
        """
        return f"GroupOfDates(dates={_shorten(self.dates)})"

    def __eq__(self, other: object) -> bool:
        """Check if two groups of dates are equal.

        Args:
            other (object): The other group of dates.

        Returns:
            bool: True if the groups are equal, False otherwise.
        """
        return isinstance(other, GroupOfDates) and self.dates == other.dates


class Groups:
    """A collection of groups of dates.

    Examples:
        >>> list(Groups(group_by="daily", start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=12))[0]
        [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 1, 12, 0)]

        >>> list(Groups(group_by="daily", start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=12))[1]
        [datetime.datetime(2023, 1, 2, 0, 0), datetime.datetime(2023, 1, 2, 12, 0)]

        >>> g = Groups(group_by=3, start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=24)
        >>> len(list(g))
        2
        >>> len(list(g)[0])
        3
        >>> len(list(g)[1])
        2
        >>> g = Groups(group_by=3,
        ...            start="2023-01-01 00:00",
        ...            end="2023-01-05 00:00",
        ...            frequency=24,
        ...            missing=["2023-01-02 00:00"])
        >>> len(list(g))
        2
        >>> len(list(g)[0])
        2
        >>> len(list(g)[1])
        2
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class with the provided keyword arguments.

        Parameters
        ----------
            **kwargs : Any : Arbitrary keyword arguments. Expected keys include:
                - group_by: Configuration for the Grouper.
                - Other keys for DatesProvider configuration.
        """

        group_by = kwargs.pop("group_by")
        self._dates = DatesProvider.from_config(**kwargs)
        self._grouper = Grouper.from_config(group_by)
        self._filter = Filter(self._dates.missing)

    @property
    def provider(self) -> DatesProvider:
        """Return the dates provider."""
        return self._dates

    def __iter__(self) -> Iterator[GroupOfDates]:
        """Return an iterator over the groups of dates.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        for go in self._grouper(self._dates):
            dates = self._filter(go.dates)
            if not dates:
                continue
            yield GroupOfDates(dates, go.provider)

    def __len__(self) -> int:
        """Return the number of groups of dates.

        Returns:
            int: The number of groups.
        """
        return self._len

    @cached_property
    def _len(self) -> int:
        """Calculate the number of groups of dates."""
        n = 0
        for go in self._grouper(self._dates):
            dates = self._filter(go.dates)
            if not dates:
                continue
            n += 1
        return n

    def __repr__(self) -> str:
        """Return a string representation of the groups of dates.

        Returns:
            str: The string representation.
        """
        return f"{self.__class__.__name__}(dates={len(self)},{_shorten(self._dates)})"

    def describe(self) -> str:
        """Return a summary description of the dates.

        Returns:
            str: The summary description.
        """
        return self._dates.summary

    def one_date(self) -> GroupOfDates:
        """Return a group containing only one date.

        Returns:
            GroupOfDates: The group containing only one date.
        """
        go = next(iter(self))
        return GroupOfDates([go.dates[0]], go.provider)


class Filter:
    """A class to filter out missing dates."""

    def __init__(self, missing: List[datetime.datetime]) -> None:
        self.missing = set(as_datetime(m) for m in missing)

    def __call__(self, dates: List[datetime.datetime]) -> List[datetime.datetime]:
        """Filter out missing dates from the list of dates.

        Args:
            dates (List[datetime.datetime]): The list of dates.

        Returns:
            List[datetime.datetime]: The filtered list of dates.
        """
        return [d for d in dates if d not in self.missing]


class Grouper(ABC):
    """Abstract base class for grouping dates."""

    @classmethod
    def from_config(cls, group_by: Any) -> "Grouper":
        """Create a grouper based on the configuration."""

        if isinstance(group_by, int) and group_by > 0:
            return GrouperByFixedSize(group_by)

        if group_by is None:
            return GrouperOneGroup()

        if group_by == "reference_date":
            return ReferenceDateGroup()

        key = {
            "monthly": lambda dt: (dt.year, dt.month),
            "daily": lambda dt: (dt.year, dt.month, dt.day),
            "weekly": lambda dt: (dt.weekday(),),
            "MMDD": lambda dt: (dt.month, dt.day),
        }[group_by]
        return GrouperByKey(key)

    @abstractmethod
    def __call__(self, dates: DatesProvider) -> Iterator[GroupOfDates]:
        """Group dates based on the implementation.

        Args:
            dates (DatesProvider): The dates provider.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        pass


class ReferenceDateGroup(Grouper):
    """Group dates by their reference date."""

    def __call__(self, dates: DatesProvider) -> Iterator[GroupOfDates]:
        """Group dates by their reference date.

        Args:
            dates (DatesProvider): The dates provider.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        assert isinstance(dates, DatesProvider), type(dates)

        mapping = dates.mapping

        def same_refdate(dt):
            return mapping[dt].refdate

        for _, g in itertools.groupby(sorted(dates, key=same_refdate), key=same_refdate):
            yield GroupOfDates(list(g), dates)


class GrouperOneGroup(Grouper):
    """Group all dates into a single group."""

    def __call__(self, dates: DatesProvider) -> Iterator[GroupOfDates]:
        """Group all dates into a single group.

        Args:
            dates (DatesProvider): The dates provider.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        assert isinstance(dates, DatesProvider), type(dates)

        yield GroupOfDates(dates.values, dates)


class GrouperByKey(Grouper):
    """Group dates by a key."""

    def __init__(self, key: Callable[[datetime.datetime], Any]) -> None:
        self.key = key

    def __call__(self, dates: DatesProvider) -> Iterator[GroupOfDates]:
        """Group dates based on the provided key.

        Args:
            dates (DatesProvider): The dates provider.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        for _, g in itertools.groupby(sorted(dates, key=self.key), key=self.key):
            yield GroupOfDates(list(g), dates)


class GrouperByFixedSize(Grouper):
    """Group dates by a fixed size."""

    def __init__(self, size: int) -> None:
        self.size = size

    def __call__(self, dates: DatesProvider) -> Iterator[GroupOfDates]:
        """Group dates into fixed-size batches.

        Args:
            dates (DatesProvider): The dates provider.

        Returns:
            Iterator[GroupOfDates]: The iterator over the groups of dates.
        """
        batch = []

        for d in dates:
            batch.append(d)
            if len(batch) == self.size:
                yield GroupOfDates(batch, dates)
                batch = []

        if batch:
            yield GroupOfDates(batch, dates)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import glob
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_grid
from anemoi.transform.fields import new_fieldlist_from_list
from anemoi.transform.flavour import RuleBasedFlavour
from anemoi.transform.grids import grid_registry
from earthkit.data import from_source
from earthkit.data.utils.patterns import Pattern

from .legacy import legacy_source

LOG = logging.getLogger(__name__)


def check(ds: Any, paths: List[str], **kwargs: Any) -> None:
    """Check if the dataset matches the expected number of fields.

    Parameters
    ----------
    ds : Any
        The dataset to check.
    paths : list of str
        List of paths to the GRIB files.
    **kwargs : Any
        Additional keyword arguments.

    Raises
    ------
    ValueError
        If the number of fields does not match the expected count.
    """
    count = 1
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list)):
            count *= len(v)

    if len(ds) != count:
        raise ValueError(f"Expected {count} fields, got {len(ds)} (kwargs={kwargs}, paths={paths})")


def _expand(paths: List[str]) -> Any:
    """Expand the given paths using glob.

    Parameters
    ----------
    paths : list of str
        List of paths to expand.

    Returns
    -------
    Any
        The expanded paths.
    """
    for path in paths:
        cnt = 0
        for p in glob.glob(path):
            yield p
            cnt += 1
        if cnt == 0:
            yield path


@legacy_source(__file__)
def execute(
    context: Any,
    dates: List[Any],
    path: Union[str, List[str]],
    flavour: Optional[Union[str, Dict[str, Any]]] = None,
    grid_definition: Optional[Dict[str, Any]] = None,
    *args: Any,
    **kwargs: Any,
) -> ekd.FieldList:
    """Execute the function to load data from GRIB files.

    Args:
        context (Any): The context in which the function is executed.
        dates (List[Any]): List of dates.
        path (Union[str, List[str]]): Path or list of paths to the GRIB files.
        flavour (Optional[Union[str, Dict[str, Any]]], optional): Flavour information. Defaults to None.
        grid_definition (Optional[Dict[str, Any]]): Grid definition config to create a Grid object.
        *args (Any): Additional arguments.
        **kwargs (Any): Additional keyword arguments.

    Returns
    -------
    Any
        The loaded dataset.
    """
    given_paths = path if isinstance(path, list) else [path]
    if flavour is not None:
        flavour = RuleBasedFlavour(flavour)

    if grid_definition is not None:
        grid = grid_registry.from_config(grid_definition)
    else:
        grid = None

    ds = from_source("empty")
    dates = [d.isoformat() for d in dates]

    for path in given_paths:
        paths = Pattern(path, ignore_missing_keys=True).substitute(*args, date=dates, **kwargs)

        for name in ("grid", "area", "rotation", "frame", "resol", "bitmap"):
            if name in kwargs:
                raise ValueError(f"MARS interpolation parameter '{name}' not supported")

        for path in _expand(paths):
            context.trace("📁", "PATH", path)
            s = from_source("file", path)
            if flavour is not None:
                s = flavour.map(s)
            s = s.sel(valid_datetime=dates, **kwargs)
            ds = ds + s

    if kwargs and not context.partial_ok:
        check(ds, given_paths, valid_datetime=dates, **kwargs)

    if grid is not None:
        ds = new_fieldlist_from_list([new_field_from_grid(f, grid) for f in ds])

    if len(ds) == 0:
        LOG.warning(f"No fields found for {dates} in {given_paths} (kwargs={kwargs})")

    return ds



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from earthkit.data import Field
from earthkit.data.core.fieldlist import math
from numpy.typing import NDArray

from .coordinates import extract_single_value
from .coordinates import is_scalar
from .metadata import XArrayMetadata

LOG = logging.getLogger(__name__)


class EmptyFieldList:
    """A class to represent an empty list of fields."""

    def __len__(self) -> int:
        """Return the length of the field list."""
        return 0

    def __getitem__(self, i: int) -> Any:
        """Raise an IndexError when trying to access an item.

        Returns
        -------
        Any
            This method does not return anything as it raises an IndexError.

        Raises
        ------
        IndexError
            Always raised to indicate that the list is empty.

        Args
        ----
        i : int
            Index of the item to access.
        """
        raise IndexError(i)

    def __repr__(self) -> str:
        """Return a string representation of the EmptyFieldList."""
        return "EmptyFieldList()"


class XArrayField(Field):
    """A class to represent a field in an XArray dataset."""

    def __init__(self, owner: Any, selection: Any) -> None:
        """Create a new XArrayField object.

        Parameters
        ----------
        owner : Variable
            The variable that owns this field.
        selection : XArrayDataArray
            A 2D sub-selection of the variable's underlying array.
            This is actually a nD object, but the first dimensions are always 1.
            The other two dimensions are latitude and longitude.
        """
        self.owner = owner
        self.selection = selection

        # Copy the metadata from the owner
        self._md = owner._metadata.copy()

        for coord_name, coord_value in self.selection.coords.items():
            if is_scalar(coord_value):
                # Extract the single value from the scalar dimension
                # and store it in the metadata
                coordinate = owner.by_name[coord_name]
                self._md[coord_name] = coordinate.normalise(extract_single_value(coord_value))

        # print(values.ndim, values.shape, selection.dims)
        # By now, the only dimensions should be latitude and longitude
        self._shape = tuple(list(self.selection.shape)[-2:])
        if math.prod(self._shape) != math.prod(self.selection.shape):
            print(self.selection.ndim, self.selection.shape)
            print(self.selection)
            raise ValueError("Invalid shape for selection")

    @property
    def shape(self) -> Tuple[int, int]:
        """Return the shape of the field."""
        return self._shape

    def to_numpy(
        self, flatten: bool = False, dtype: Optional[type] = None, index: Optional[int] = None
    ) -> NDArray[Any]:
        """Convert the selection to a numpy array.

        Returns
        -------
        NDArray[Any]
            The selection converted to a numpy array.

        Args
        ----
        flatten : bool, optional
            Whether to flatten the array, by default False.
        dtype : Optional[type], optional
            Data type of the array, by default None.
        index : Optional[int], optional
            Index to select a specific element, by default None.
        """
        if index is not None:
            values = self.selection[index]
        else:
            values = self.selection

        assert dtype is None

        if flatten:
            return values.values.flatten()

        return values  # .reshape(self.shape)

    @cached_property
    def _metadata(self) -> XArrayMetadata:
        """Return the metadata of the field."""
        return XArrayMetadata(self)

    def grid_points(self) -> Any:
        """Return the grid points of the field."""
        return self.owner.grid_points()

    def to_latlon(self, flatten: bool = True) -> Dict[str, Any]:
        """Convert the selection to latitude and longitude coordinates.

        Returns
        -------
        Dict[str, Any]
            The latitude and longitude coordinates.

        Args
        ----
        flatten : bool, optional
            Whether to flatten the coordinates, by default True.
        """
        assert flatten
        return dict(lat=self.latitudes, lon=self.longitudes)

    @property
    def resolution(self) -> Optional[Any]:
        """Return the resolution of the field."""
        return None

    @property
    def grid_mapping(self) -> Any:
        """Return the grid mapping of the field."""
        return self.owner.grid_mapping

    @property
    def latitudes(self) -> Any:
        """Return the latitudes of the field."""
        return self.owner.latitudes

    @property
    def longitudes(self) -> Any:
        """Return the longitudes of the field."""
        return self.owner.longitudes

    @property
    def forecast_reference_time(self) -> datetime.datetime:
        """Return the forecast reference time of the field."""
        date, time = self.metadata("date", "time")
        assert len(time) == 4, time
        assert len(date) == 8, date
        yyyymmdd = int(date)
        time = int(time) // 100
        return datetime.datetime(yyyymmdd // 10000, yyyymmdd // 100 % 100, yyyymmdd % 100, time)

    def __repr__(self) -> str:
        """Return a string representation of the field."""
        return repr(self._metadata)

    def _values(self, dtype: Optional[type] = None) -> Any:
        """Return the values of the selection.

        Returns
        -------
        Any
            The values of the selection.

        Args
        ----
        dtype : Optional[type], optional
            Data type of the values, by default None.
        """
        # we don't use .values as this will download the data
        return self.selection



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os

from anemoi.utils.testing import get_test_data
from anemoi.utils.testing import skip_if_offline
from anemoi.utils.testing import skip_missing_packages
from anemoi.utils.testing import skip_slow_tests

from anemoi.datasets import open_dataset
from anemoi.datasets.create.testing import create_dataset


@skip_if_offline
def test_grib() -> None:
    """Test the creation of a dataset from GRIB files.

    This function tests the creation of a dataset using GRIB files from
    specific dates and verifies the shape of the resulting dataset.
    """
    data1 = get_test_data("anemoi-datasets/create/grib-20100101.grib")
    data2 = get_test_data("anemoi-datasets/create/grib-20100102.grib")
    assert os.path.dirname(data1) == os.path.dirname(data2)

    path = os.path.dirname(data1)

    config = {
        "dates": {
            "start": "2010-01-01T00:00:00",
            "end": "2010-01-02T18:00:00",
            "frequency": "6h",
        },
        "input": {
            "grib": {
                "path": os.path.join(path, "grib-{date:strftime(%Y%m%d)}.grib"),
            },
        },
    }

    created = create_dataset(config=config, output=None)
    ds = open_dataset(created)
    assert ds.shape == (8, 12, 1, 162)


@skip_if_offline
def test_netcdf() -> None:
    """Test for NetCDF files.

    This function tests the creation of a dataset from a NetCDF file.
    """
    data = get_test_data("anemoi-datasets/create/netcdf.nc")
    config = {
        "dates": {
            "start": "2023-01-01",
            "end": "2023-01-02",
            "frequency": "1d",
        },
        "input": {
            "netcdf": {"path": data},
        },
    }

    created = create_dataset(config=config, output=None)
    ds = open_dataset(created)
    assert ds.shape == (2, 2, 1, 162)


@skip_missing_packages("fstd", "rpnpy.librmn")
def test_eccs_fstd() -> None:
    """Test for 'fstd' files from ECCC."""
    # See https://github.com/neishm/fstd2nc

    data = get_test_data("anemoi-datasets/create/2025031000_000_TT.fstd", gzipped=True)
    config = {
        "dates": {
            "start": "2023-01-01",
            "end": "2023-01-02",
            "frequency": "1d",
        },
        "input": {
            "eccc_fstd": {"path": data},
        },
    }

    created = create_dataset(config=config, output=None)
    ds = open_dataset(created)
    assert ds.shape == (2, 2, 1, 162)


@skip_slow_tests
@skip_if_offline
@skip_missing_packages("kerchunk", "s3fs", "h5py")
def test_kerchunk() -> None:
    """Test for Kerchunk JSON files.

    This function tests the creation of a dataset from a Kerchunk JSON file.

    """
    # Note: last version of kerchunk compatible with zarr 2 is 0.2.7

    data = get_test_data("anemoi-datasets/create/kerchunck.json", gzipped=True)

    config = {
        "dates": {
            "start": "2024-03-01T00:00:00",
            "end": "2024-03-01T18:00:00",
            "frequency": "6h",
        },
        "input": {
            "xarray-kerchunk": {
                "json": data,
                "param": ["T"],
                "level": [1000],
            },
        },
    }

    created = create_dataset(config=config, output=None)
    ds = open_dataset(created)
    assert ds.shape == (4, 1, 1, 1038240)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    test_kerchunk()
    exit()
    """Run all test functions that start with 'test_'."""
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import json
import logging
import pprint
import warnings
from abc import ABC
from abc import abstractmethod
from functools import cached_property

try:
    from types import EllipsisType
except ImportError:
    # Python 3.9
    EllipsisType = type(Ellipsis)
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Sized
from typing import Tuple
from typing import Union

import numpy as np
from anemoi.utils.dates import frequency_to_seconds
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta
from numpy.typing import NDArray

from .debug import Node
from .debug import Source

if TYPE_CHECKING:
    import matplotlib

LOG = logging.getLogger(__name__)


Shape = Tuple[int, ...]
TupleIndex = Tuple[Union[int, slice, EllipsisType], ...]
FullIndex = Union[int, slice, TupleIndex]


def _tidy(v: Any) -> Any:
    """Tidy up the input value.

    Parameters
    ----------
    v : Any
        The input value to tidy up.

    Returns
    -------
    Any
        The tidied value.
    """
    if isinstance(v, (list, tuple, set)):
        return [_tidy(i) for i in v]
    if isinstance(v, dict):
        return {k: _tidy(v) for k, v in v.items()}
    if isinstance(v, datetime.datetime):
        return v.isoformat()
    if isinstance(v, datetime.date):
        return v.isoformat()
    if isinstance(v, datetime.timedelta):
        return frequency_to_string(v)

    if isinstance(v, Dataset):
        # That can happen in the `arguments`
        # if a dataset is passed as an argument
        return repr(v)

    if isinstance(v, slice):
        return (v.start, v.stop, v.step)

    if isinstance(v, np.integer):
        return int(v)

    return v


class Dataset(ABC, Sized):
    arguments: Dict[str, Any] = {}
    _name: Union[str, None] = None

    def mutate(self) -> "Dataset":
        """Give an opportunity to a subclass to return a new Dataset object of a different class, if needed.

        Returns
        -------
        Dataset
            The mutated dataset.
        """
        return self

    def swap_with_parent(self, parent: "Dataset") -> "Dataset":
        """Swap the current dataset with its parent dataset.

        Parameters
        ----------
        parent : Dataset
            The parent dataset.

        Returns
        -------
        Dataset
            The parent dataset.
        """
        return parent

    @cached_property
    def _len(self) -> int:
        """Cache and return the length of the dataset."""
        return len(self)

    def _subset(self, **kwargs: Any) -> "Dataset":
        """Create a subset of the dataset based on the provided keyword arguments.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments for creating the subset.

        Returns
        -------
        Dataset
            The subset of the dataset.
        """
        if not kwargs:
            return self.mutate()

        name = kwargs.pop("name", None)
        result = self.__subset(**kwargs)
        result._name = name

        return result

    @property
    def name(self) -> Union[str, None]:
        """Return the name of the dataset."""
        return self._name

    def __subset(self, **kwargs: Any) -> "Dataset":
        """Internal method to create a subset of the dataset based on the provided keyword arguments.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments for creating the subset.

        Returns
        -------
        Dataset
            The subset of the dataset.
        """
        if not kwargs:
            return self.mutate()

        # This one must be first
        if "fill_missing_dates" in kwargs:
            from .fill_missing import fill_missing_dates_factory

            fill_missing_dates = kwargs.pop("fill_missing_dates")
            ds = fill_missing_dates_factory(self, fill_missing_dates, kwargs)
            return ds._subset(**kwargs).mutate()

        if "start" in kwargs or "end" in kwargs:
            start = kwargs.pop("start", None)
            end = kwargs.pop("end", None)

            from .subset import Subset

            return (
                Subset(self, self._dates_to_indices(start, end), dict(start=start, end=end))._subset(**kwargs).mutate()
            )

        if "frequency" in kwargs:
            from .subset import Subset

            if "interpolate_frequency" in kwargs:
                raise ValueError("Cannot use both `frequency` and `interpolate_frequency`")

            frequency = kwargs.pop("frequency")
            return (
                Subset(self, self._frequency_to_indices(frequency), dict(frequency=frequency))
                ._subset(**kwargs)
                .mutate()
            )

        if "select" in kwargs:
            from .select import Select

            select = kwargs.pop("select")
            return Select(self, self._select_to_columns(select), {"select": select})._subset(**kwargs).mutate()

        if "drop" in kwargs:
            from .select import Select

            drop = kwargs.pop("drop")
            return Select(self, self._drop_to_columns(drop), {"drop": drop})._subset(**kwargs).mutate()

        if "reorder" in kwargs:
            from .select import Select

            reorder = kwargs.pop("reorder")
            return Select(self, self._reorder_to_columns(reorder), {"reoder": reorder})._subset(**kwargs).mutate()

        if "rename" in kwargs:
            from .select import Rename

            rename = kwargs.pop("rename")
            return Rename(self, rename)._subset(**kwargs).mutate()

        if "rescale" in kwargs:
            from .rescale import Rescale

            rescale = kwargs.pop("rescale")
            return Rescale(self, rescale)._subset(**kwargs).mutate()

        if "statistics" in kwargs:
            from ..data import open_dataset
            from .statistics import Statistics

            statistics = kwargs.pop("statistics")

            return Statistics(self, open_dataset(statistics))._subset(**kwargs).mutate()

        # Note: trim_edge should go before thinning
        if "trim_edge" in kwargs:
            from .masked import TrimEdge

            edge = kwargs.pop("trim_edge")
            return TrimEdge(self, edge)._subset(**kwargs).mutate()

        if "thinning" in kwargs:
            from .masked import Thinning

            thinning = kwargs.pop("thinning")
            method = kwargs.pop("method", "every-nth")
            return Thinning(self, thinning, method)._subset(**kwargs).mutate()

        if "area" in kwargs:
            from .masked import Cropping

            bbox = kwargs.pop("area")
            return Cropping(self, bbox)._subset(**kwargs).mutate()

        if "number" in kwargs or "numbers" in kwargs or "member" in kwargs or "members" in kwargs:
            from .ensemble import Number

            members = {}
            for key in ["number", "numbers", "member", "members"]:
                if key in kwargs:
                    members[key] = kwargs.pop(key)

            return Number(self, **members)._subset(**kwargs).mutate()

        if "set_missing_dates" in kwargs:
            from .missing import MissingDates

            set_missing_dates = kwargs.pop("set_missing_dates")
            return MissingDates(self, set_missing_dates)._subset(**kwargs).mutate()

        if "skip_missing_dates" in kwargs:
            from .missing import SkipMissingDates

            if "expected_access" not in kwargs:
                raise ValueError("`expected_access` is required with `skip_missing_dates`")

            skip_missing_dates = kwargs.pop("skip_missing_dates")
            expected_access = kwargs.pop("expected_access")

            if skip_missing_dates:
                return SkipMissingDates(self, expected_access)._subset(**kwargs).mutate()

        if "interpolate_frequency" in kwargs:
            from .interpolate import InterpolateFrequency

            interpolate_frequency = kwargs.pop("interpolate_frequency")
            return InterpolateFrequency(self, interpolate_frequency)._subset(**kwargs).mutate()

        if "interpolate_variables" in kwargs:
            from .interpolate import InterpolateNearest

            interpolate_variables = kwargs.pop("interpolate_variables")
            max_distance = kwargs.pop("max_distance", None)
            return InterpolateNearest(self, interpolate_variables, max_distance=max_distance)._subset(**kwargs).mutate()

        # Keep last
        if "shuffle" in kwargs:
            from .subset import Subset

            shuffle = kwargs.pop("shuffle")

            if shuffle:
                return Subset(self, self._shuffle_indices(), dict(shuffle=True))._subset(**kwargs).mutate()

        raise NotImplementedError("Unsupported arguments: " + ", ".join(kwargs))

    def _frequency_to_indices(self, frequency: str) -> list[int]:
        """Convert a frequency string to a list of indices.

        Parameters
        ----------
        frequency : str
            The frequency string.

        Returns
        -------
        list of int
            The list of indices.
        """
        requested_frequency = frequency_to_seconds(frequency)
        dataset_frequency = frequency_to_seconds(self.frequency)

        if requested_frequency % dataset_frequency != 0:
            raise ValueError(
                f"Requested frequency {frequency} is not a multiple of the dataset frequency {self.frequency}. Did you mean to use `interpolate_frequency`?"
            )

        # Question: where do we start? first date, or first date that is a multiple of the frequency?
        step = requested_frequency // dataset_frequency

        return range(0, len(self), step)

    def _shuffle_indices(self) -> NDArray[Any]:
        """Return a shuffled array of indices.

        Returns
        -------
        numpy.ndarray
            The shuffled array of indices.
        """
        return np.random.permutation(len(self))

    def _dates_to_indices(
        self,
        start: Union[None, str, datetime.datetime],
        end: Union[None, str, datetime.datetime],
    ) -> List[int]:
        """Convert date range to a list of indices.

        Parameters
        ----------
        start : None, str, or datetime.datetime
            The start date.
        end : None, str, or datetime.datetime
            The end date.

        Returns
        -------
        list of int
            The list of indices.
        """
        from .misc import as_first_date
        from .misc import as_last_date

        # TODO: optimize

        start = self.dates[0] if start is None else as_first_date(start, self.dates)
        end = self.dates[-1] if end is None else as_last_date(end, self.dates)

        return [i for i, date in enumerate(self.dates) if start <= date <= end]

    def _select_to_columns(self, vars: Union[str, List[str], Tuple[str], set]) -> List[int]:
        """Convert variable names to a list of column indices.

        Parameters
        ----------
        vars : str, list of str, tuple of str, or set
            The variable names.

        Returns
        -------
        list of int
            The list of column indices.
        """
        if isinstance(vars, set):
            # We keep the order of the variables as they are in the zarr file
            nvars = [v for v in self.name_to_index if v in vars]
            assert len(nvars) == len(vars)
            return self._select_to_columns(nvars)

        if not isinstance(vars, (list, tuple)):
            vars = [vars]

        return [self.name_to_index[v] for v in vars]

    def _drop_to_columns(self, vars: Union[str, Sequence[str]]) -> List[int]:
        """Convert variable names to a list of column indices to drop.

        Parameters
        ----------
        vars : str, list of str, tuple of str, or set
            The variable names.

        Returns
        -------
        list of int
            The list of column indices to drop.
        """
        if not isinstance(vars, (list, tuple, set)):
            vars = [vars]

        if not set(vars) <= set(self.name_to_index):
            raise ValueError(f"drop: unknown variables: {set(vars) - set(self.name_to_index)}")

        return sorted([v for k, v in self.name_to_index.items() if k not in vars])

    def _reorder_to_columns(self, vars: Union[str, List[str], Tuple[str], Dict[str, int]]) -> List[int]:
        """Convert variable names to a list of reordered column indices.

        Parameters
        ----------
        vars : str, list of str, tuple of str, or dict of str to int
            The variable names.

        Returns
        -------
        list of int
            The list of reordered column indices.
        """
        if isinstance(vars, str) and vars == "sort":
            # Sorting the variables alphabetically.
            # This is cruical for pre-training then transfer learning in combination with
            # cutout and adjust = 'all'

            indices = [self.name_to_index[k] for k, v in sorted(self.name_to_index.items(), key=lambda x: x[0])]
            assert set(indices) == set(range(len(self.name_to_index)))
            return indices

        if isinstance(vars, (list, tuple)):
            vars = {k: i for i, k in enumerate(vars)}

        indices = [self.name_to_index[k] for k, v in sorted(vars.items(), key=lambda x: x[1])]

        # Make sure we don't forget any variables
        assert set(indices) == set(range(len(self.name_to_index)))

        return indices

    def dates_interval_to_indices(
        self, start: Union[None, str, datetime.datetime], end: Union[None, str, datetime.datetime]
    ) -> List[int]:
        """Convert date interval to a list of indices.

        Parameters
        ----------
        start : None, str, or datetime.datetime
            The start date.
        end : None, str, or datetime.datetime
            The end date.

        Returns
        -------
        list of int
            The list of indices.
        """
        return self._dates_to_indices(start, end)

    def provenance(self) -> Dict[str, Any]:
        """Return the provenance information of the dataset.

        Returns
        -------
        dict
            The provenance information.
        """
        return {}

    def sub_shape(self, drop_axis: int) -> TupleIndex:
        """Return the shape of the dataset with one axis dropped.

        Parameters
        ----------
        drop_axis : int
            The axis to drop.

        Returns
        -------
        tuple
            The shape with one axis dropped.
        """
        shape = list(self.shape)
        shape.pop(drop_axis)
        return tuple(shape)

    @property
    def typed_variables(self) -> Dict[str, Any]:
        """Return the variables with their types."""
        from anemoi.transform.variables import Variable

        constants = self.constant_fields

        result = {}
        for k, v in self.variables_metadata.items():

            # TODO: Once all datasets are updated, we can remove this
            v = v.copy()
            if k in constants:
                v["constant_in_time"] = True

            if "is_constant_in_time" in v:
                del v["is_constant_in_time"]

            result[k] = Variable.from_dict(k, v)

        return result

    def _input_sources(self) -> List[Any]:
        """Return the input sources of the dataset.

        Returns
        -------
        list
            The input sources.
        """
        sources = []
        self.collect_input_sources(sources)
        return sources

    def metadata(self) -> Dict[str, Any]:
        """Return the metadata of the dataset.

        Returns
        -------
        dict
            The metadata.
        """
        import anemoi

        _, source_to_arrays = self._supporting_arrays_and_sources()

        sources = []
        for i, source in enumerate(self._input_sources()):
            source_metadata = source.dataset_metadata().copy()
            source_metadata["supporting_arrays"] = source_to_arrays[id(source)]
            sources.append(source_metadata)

        md = dict(
            version=anemoi.datasets.__version__,
            arguments=self.arguments,
            **self.dataset_metadata(),
            sources=sources,
            supporting_arrays=source_to_arrays[id(self)],
        )

        try:
            return json.loads(json.dumps(_tidy(md)))
        except Exception:
            LOG.exception("Failed to serialize metadata")
            pprint.pprint(md)

            raise

    @property
    def start_date(self) -> np.datetime64:
        """Return the start date of the dataset."""
        return self.dates[0]

    @property
    def end_date(self) -> np.datetime64:
        """Return the end date of the dataset."""
        return self.dates[-1]

    def dataset_metadata(self) -> Dict[str, Any]:
        """Return the metadata of the dataset.

        Returns
        -------
        dict
            The metadata.
        """
        return dict(
            specific=self.metadata_specific(),
            frequency=self.frequency,
            variables=self.variables,
            variables_metadata=self.variables_metadata,
            shape=self.shape,
            dtype=str(self.dtype),
            start_date=self.start_date.astype(str),
            end_date=self.end_date.astype(str),
            name=self.name,
        )

    def _supporting_arrays(self, *path: str) -> Dict[str, NDArray[Any]]:
        """Return the supporting arrays of the dataset.

        Parameters
        ----------
        *path : str
            The path components.

        Returns
        -------
        dict
            The supporting arrays.
        """

        def _path(path, name: str) -> str:
            return "/".join(str(_) for _ in [*path, name])

        result = {
            _path(path, "latitudes"): self.latitudes,
            _path(path, "longitudes"): self.longitudes,
        }
        collected = []

        self.collect_supporting_arrays(collected, *path)

        for path, name, array in collected:
            assert isinstance(path, tuple) and isinstance(name, str)
            assert isinstance(array, np.ndarray)

            name = _path(path, name)

            if name in result:
                raise ValueError(f"Duplicate key {name}")

            result[name] = array

        return result

    def supporting_arrays(self) -> Dict[str, NDArray[Any]]:
        """Return the supporting arrays to be saved in the checkpoints.

        Returns
        -------
        dict
            The supporting arrays.
        """
        arrays, _ = self._supporting_arrays_and_sources()
        return arrays

    def _supporting_arrays_and_sources(self) -> Tuple[Dict[str, NDArray], Dict[int, List[str]]]:
        """Return the supporting arrays and their sources.

        Returns
        -------
        tuple
            The supporting arrays and their sources.
        """
        source_to_arrays = {}

        # Top levels arrays
        result = self._supporting_arrays()
        source_to_arrays[id(self)] = sorted(result.keys())

        # Arrays from the input sources
        for i, source in enumerate(self._input_sources()):
            name = source.name if source.name is not None else f"source{i}"
            src_arrays = source._supporting_arrays(name)
            source_to_arrays[id(source)] = sorted(src_arrays.keys())

            for k in src_arrays:
                assert k not in result

            result.update(src_arrays)

        return result, source_to_arrays

    def collect_supporting_arrays(self, collected: List[Tuple[Tuple[str, ...], str, NDArray[Any]]], *path: str) -> None:
        """Collect supporting arrays.

        Parameters
        ----------
        collected : list of tuple
            The collected supporting arrays.
        *path : str
            The path components.
        """
        # Override this method to add more arrays
        pass

    def metadata_specific(self, **kwargs: Any) -> Dict[str, Any]:
        """Return specific metadata of the dataset.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        dict
            The specific metadata.
        """
        action = self.__class__.__name__.lower()
        # assert isinstance(self.frequency, datetime.timedelta), (self.frequency, self, action)
        return dict(
            action=action,
            variables=self.variables,
            shape=self.shape,
            frequency=frequency_to_string(frequency_to_timedelta(self.frequency)),
            start_date=self.start_date.astype(str),
            end_date=self.end_date.astype(str),
            **kwargs,
        )

    def __repr__(self) -> str:
        """Return the string representation of the dataset.

        Returns
        -------
        str
            The string representation.
        """
        return self.__class__.__name__ + "()"

    @property
    def grids(self) -> TupleIndex:
        """Return the grid shape of the dataset."""
        return (self.shape[-1],)

    def _check(self) -> None:
        """Check for overridden private methods in the dataset."""
        common = Dataset.__dict__.keys() & self.__class__.__dict__.keys()
        overriden = [m for m in common if Dataset.__dict__[m] is not self.__class__.__dict__[m]]

        for n in overriden:
            if n.startswith("_") and not n.startswith("__") and n not in ("_abc_impl",):
                warnings.warn(f"Private method {n} is overriden in {self.__class__.__name__}")

    def _repr_html_(self) -> str:
        """Return the HTML representation of the dataset.

        Returns
        -------
        str
            The HTML representation.
        """
        return self.tree().html()

    @property
    def label(self) -> str:
        """Return the label of the dataset."""
        return self.__class__.__name__.lower()

    def computed_constant_fields(self) -> List[str]:
        """Return the computed constant fields of the dataset.

        Returns
        -------
        list of str
            The computed constant fields.
        """
        try:
            # If the tendencies are computed, we can use them
            return sorted(self._compute_constant_fields_from_statistics())
        except KeyError:
            # This can happen if the tendencies are not computed
            pass

        return sorted(self._compute_constant_fields_from_a_few_samples())

    def _compute_constant_fields_from_a_few_samples(self) -> List[str]:
        """Compute constant fields from a few samples.

        Returns
        -------
        list of str
            The computed constant fields.
        """
        import numpy as np

        # Otherwise, we need to compute them
        dates = self.dates
        indices = set(range(len(dates)))
        indices -= self.missing

        sample_count = min(4, len(indices))
        count = len(indices)

        p = slice(0, count, count // max(1, sample_count - 1))
        samples = list(range(*p.indices(count)))

        samples.append(count - 1)  # Add last
        samples = sorted(set(samples))
        indices = list(indices)
        samples = [indices[i] for i in samples]

        assert set(samples) <= set(indices)  # Make sure we have the samples

        first = None
        constants = [True] * len(self.variables)

        first = self[samples.pop(0)]

        for sample in samples:
            row = self[sample]
            for i, (a, b) in enumerate(zip(row, first)):
                if np.any(a != b):
                    constants[i] = False

        return [v for i, v in enumerate(self.variables) if constants[i]]

    def _compute_constant_fields_from_statistics(self) -> List[str]:
        """Compute constant fields from statistics.

        Returns
        -------
        list of str
            The computed constant fields.
        """
        result = []

        t = self.statistics_tendencies()

        for i, v in enumerate(self.variables):
            if t["mean"][i] == 0 and t["stdev"][i] == 0:
                result.append(v)

        return result

    def plot(
        self,
        date: Union[int, datetime.datetime, np.datetime64, str],
        variable: Union[int, str],
        member: int = 0,
        **kwargs: Any,
    ) -> "matplotlib.pyplot.Axes":
        """For debugging purposes, plot a field.

        Parameters
        ----------
        date : int or datetime.datetime or numpy.datetime64 or str
            The date to plot.
        variable : int or str
            The variable to plot.
        member : int, optional
            The ensemble member to plot.
        **kwargs : Any
            Additional arguments to pass to matplotlib.pyplot.tricontourf.

        Returns
        -------
        matplotlib.pyplot.Axes
            The plot axes.
        """
        from anemoi.utils.devtools import plot_values

        values = self[self.to_index(date, variable, member)]

        return plot_values(values, self.latitudes, self.longitudes, **kwargs)

    def to_index(
        self,
        date: Union[int, datetime.datetime, np.datetime64, str],
        variable: Union[int, str],
        member: int = 0,
    ) -> Tuple[int, int, int]:
        """Convert date, variable, and member to indices.

        Parameters
        ----------
        date : int or datetime.datetime or numpy.datetime64 or str
            The date.
        variable : int or str
            The variable.
        member : int, optional
            The ensemble member.

        Returns
        -------
        tuple of int
            The indices.
        """
        from earthkit.data.utils.dates import to_datetime

        if not isinstance(date, int):
            date = np.datetime64(to_datetime(date)).astype(self.dates[0].dtype)
            index = np.where(self.dates == date)[0]
            if len(index) == 0:
                raise ValueError(
                    f"Date {date} not found in the dataset {self.dates[0]} to {self.dates[-1]} by {self.frequency}"
                )
            date_index = index[0]
        else:
            date_index = date

        date_index = int(date_index)  # because np.int64 is not instance of int

        if isinstance(variable, int):
            variable_index = variable
        else:
            if variable not in self.variables:
                raise ValueError(f"Unknown variable {variable} (available: {self.variables})")

            variable_index = self.name_to_index[variable]

        return (date_index, variable_index, member)

    @abstractmethod
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get the item at the specified index.

        Parameters
        ----------
        n : FullIndex
            Index to retrieve.

        Returns
        -------
        NDArray[Any]
            Retrieved item.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Return the length of the dataset.

        Returns
        -------
        int
            The length of the dataset.
        """

    @property
    @abstractmethod
    def variables(self) -> List[str]:
        """Return the list of variables in the dataset."""
        pass

    @property
    @abstractmethod
    def frequency(self) -> datetime.timedelta:
        """Return the frequency of the dataset."""
        pass

    @property
    @abstractmethod
    def dates(self) -> NDArray[np.datetime64]:
        """Return the dates in the dataset."""
        pass

    @property
    @abstractmethod
    def resolution(self) -> str:
        """Return the resolution of the dataset."""
        pass

    @property
    @abstractmethod
    def name_to_index(self) -> Dict[str, int]:
        """Return the mapping of variable names to indices."""
        pass

    @property
    @abstractmethod
    def shape(self) -> Shape:
        """Return the shape of the dataset."""
        pass

    @property
    @abstractmethod
    def field_shape(self) -> Shape:
        """Return the shape of the fields in the dataset."""
        pass

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        """Return the data type of the dataset."""
        pass

    @property
    @abstractmethod
    def latitudes(self) -> NDArray[Any]:
        """Return the latitudes in the dataset."""
        pass

    @property
    @abstractmethod
    def longitudes(self) -> NDArray[Any]:
        """Return the longitudes in the dataset."""
        pass

    @property
    @abstractmethod
    def variables_metadata(self) -> Dict[str, Any]:
        """Return the metadata of the variables in the dataset."""
        pass

    @abstractmethod
    @cached_property
    def missing(self) -> Set[int]:
        """Return the set of missing indices in the dataset."""
        pass

    @abstractmethod
    @cached_property
    def constant_fields(self) -> List[str]:
        """Return the list of constant fields in the dataset."""
        pass

    @abstractmethod
    @cached_property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Return the statistics of the dataset."""
        pass

    @abstractmethod
    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Return the tendencies of the statistics in the dataset.

        Parameters
        ----------
        delta : datetime.timedelta, optional
            The time delta for computing tendencies.

        Returns
        -------
        dict
            The tendencies.
        """
        pass

    @abstractmethod
    def source(self, index: int) -> Source:
        """Return the source of the dataset at the specified index.

        Parameters
        ----------
        index : int
            The index.

        Returns
        -------
        Source
            The source.
        """
        pass

    @abstractmethod
    def tree(self) -> Node:
        """Return the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation.
        """
        pass

    @abstractmethod
    def collect_input_sources(self, sources: List[Any]) -> None:
        """Collect the input sources of the dataset.

        Parameters
        ----------
        sources : list
            The input sources.
        """
        pass

    @abstractmethod
    def get_dataset_names(self, names: Set[str]) -> None:
        """Get the names of the datasets.

        Parameters
        ----------
        names : set of str
            The dataset names.
        """
        pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from anemoi.datasets.create.utils import to_datetime
from anemoi.datasets.data import MissingDateError

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import expand_list_indexing
from .indexing import update_tuple

LOG = logging.getLogger(__name__)

# TODO: Use that class instead of ZarrMissing


class MissingDates(Forwards):
    """Handles missing dates in a dataset.

    Attributes
    ----------
    dataset : Dataset
        The dataset object.
    missing_dates : List[Union[int, str]]
        List of missing dates.
    """

    def __init__(self, dataset: Dataset, missing_dates: List[Union[int, str]]) -> None:
        """Initializes the MissingDates class.

        Parameters
        ----------
        dataset : Dataset
            The dataset object.
        missing_dates : List[Union[int, str]]
            List of missing dates.
        """
        super().__init__(dataset)
        self.missing_dates = []

        self._missing = set()

        other = []
        for date in missing_dates:
            if isinstance(date, int):
                self._missing.add(date)
                self.missing_dates.append(dataset.dates[date])
            else:
                date = to_datetime(date)
                other.append(date)

        if other:
            for i, date in enumerate(dataset.dates):
                if date in other:
                    self._missing.add(i)
                    self.missing_dates.append(date)

        n = self.forward._len
        self._missing = set(i for i in self._missing if 0 <= i < n)
        self.missing_dates = sorted(to_datetime(x) for x in self.missing_dates)

        assert len(self._missing), "No dates to force missing"

    @cached_property
    def missing(self) -> Set[int]:
        """Returns the set of missing indices."""
        return self._missing.union(self.forward.missing)

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieves the item at the given index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve.

        Returns
        -------
        NDArray[Any]
            The item at the given index.
        """
        if isinstance(n, int):
            if n in self.missing:
                self._report_missing(n)
            return self.forward[n]

        if isinstance(n, slice):
            common = set(range(*n.indices(len(self)))) & self.missing
            if common:
                self._report_missing(list(common)[0])
            return self.forward[n]

        if isinstance(n, tuple):
            first = n[0]
            if isinstance(first, int):
                if first in self.missing:
                    self._report_missing(first)
                return self.forward[n]

            if isinstance(first, slice):
                common = set(range(*first.indices(len(self)))) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.forward[n]

            if isinstance(first, (list, tuple)):
                common = set(first) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.forward[n]

        raise TypeError(f"Unsupported index {n} {type(n)}")

    def _report_missing(self, n: int) -> None:
        """Reports a missing date.

        Parameters
        ----------
        n : int
            The index of the missing date.
        """
        raise MissingDateError(f"Date {self.forward.dates[n]} is missing (index={n})")

    @property
    def reason(self) -> Dict[str, Any]:
        """Provides the reason for missing dates."""
        return {"missing_dates": self.missing_dates}

    def tree(self) -> Node:
        """Builds a tree representation of the missing dates.

        Returns
        -------
        Node
            The tree representation of the missing dates.
        """
        return Node(self, [self.forward.tree()], **self.reason)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Provides metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the subclass.
        """
        return {"missing_dates": self.missing_dates}


class SkipMissingDates(Forwards):
    """Skips missing dates in a dataset.

    Attributes
    ----------
    dataset : Dataset
        The dataset object.
    expected_access : Union[int, slice]
        The expected access pattern.
    """

    def __init__(self, dataset: Dataset, expected_access: Union[int, slice]) -> None:
        """Initializes the SkipMissingDates class.

        Parameters
        ----------
        dataset : Dataset
            The dataset object.
        expected_access : Union[int, slice]
            The expected access pattern.
        """
        super().__init__(dataset)

        # if isinstance(expected_access, (tuple, list)):
        #     expected_access = slice(*expected_access)

        if isinstance(expected_access, int):
            expected_access = slice(0, expected_access)

        assert isinstance(expected_access, slice), f"Expected access must be a slice, got {expected_access}"

        expected_access = slice(*expected_access.indices(dataset._len))
        missing = dataset.missing.copy()

        size = (expected_access.stop - expected_access.start) // expected_access.step
        indices = []

        for i in range(dataset._len):
            s = slice(expected_access.start + i, expected_access.stop + i, expected_access.step)
            p = set(range(*s.indices(dataset._len)))
            if p.intersection(missing):
                continue

            if len(p) != size:
                continue

            indices.append(tuple(sorted(p)))

        self.expected_access = expected_access
        self.indices = indices

    def __len__(self) -> int:
        """Returns the length of the indices.

        Returns
        -------
        int
            The length of the indices.
        """
        return len(self.indices)

    @property
    def start_date(self) -> np.datetime64:
        """Returns the start date."""
        return self.forward.start_date

    @property
    def end_date(self) -> np.datetime64:
        """Returns the end date."""
        return self.forward.end_date

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Not implemented. Raises an error."""
        raise NotImplementedError("SkipMissingDates.dates")

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Retrieves a tuple of items at the given index.

        Parameters
        ----------
        index : TupleIndex
            The index to retrieve.

        Returns
        -------
        NDArray[Any]
            The tuple of items at the given index.
        """

        def _get_one(n):
            result = []
            for i in self.indices[n]:
                s, _ = update_tuple(index, 0, i)
                result.append(self.forward[s])

            return tuple(result)

        first = index[0]
        if isinstance(first, int):
            return _get_one(first)

        assert isinstance(first, slice), f"SkipMissingDates._get_tuple {index}"

        values = [_get_one(i) for i in range(*first.indices(self._len))]

        result = [_ for _ in zip(*values)]
        return tuple(np.stack(_) for _ in result)

    @debug_indexing
    def _get_slice(self, s: slice) -> Tuple[NDArray[Any], ...]:
        """Retrieves a slice of items.

        Parameters
        ----------
        s : slice
            The slice to retrieve.

        Returns
        -------
        Tuple[NDArray[Any], ...]
            The slice of items.
        """
        values = [self[i] for i in range(*s.indices(self._len))]
        result = [_ for _ in zip(*values)]
        return tuple(np.stack(_) for _ in result)

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> Tuple[NDArray[Any], ...]:
        """Retrieves the item at the given index.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve.

        Returns
        -------
        Tuple[NDArray[Any], ...]
            The item at the given index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return tuple(self.forward[i] for i in self.indices[n])

    @property
    def frequency(self) -> datetime.timedelta:
        """Returns the frequency of the dataset."""
        return self.forward.frequency

    def tree(self) -> Node:
        """Builds a tree representation of the skipped missing dates.

        Returns
        -------
        Node
            The tree representation of the skipped missing dates.
        """
        return Node(self, [self.forward.tree()], expected_access=self.expected_access)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Provides metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the subclass.
        """
        return {"expected_access": self.expected_access}


class MissingDataset(Forwards):
    """Represents a dataset with missing dates.

    Attributes
    ----------
    dataset : Dataset
        The dataset object.
    start : np.datetime64
        The start date.
    end : np.datetime64
        The end date.
    """

    def __init__(self, dataset: Dataset, start: np.datetime64, end: np.datetime64) -> None:
        """Initializes the MissingDataset class.

        Parameters
        ----------
        dataset : Dataset
            The dataset object.
        start : np.datetime64
            The start date.
        end : np.datetime64
            The end date.
        """
        super().__init__(dataset)
        self.start = start
        self.end = end

        dates = []
        date = start
        while date <= end:
            dates.append(date)
            date += dataset.frequency

        self._dates = np.array(dates, dtype="datetime64")
        self._missing = set(range(len(dates)))

    def __len__(self) -> int:
        """Returns the length of the dates.

        Returns
        -------
        int
            The length of the dates.
        """
        return len(self._dates)

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Returns the dates of the dataset."""
        return self._dates

    @property
    def missing(self) -> Set[int]:
        """Returns the set of missing indices."""
        return self._missing

    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Raises an error for missing dates.

        Parameters
        ----------
        n : FullIndex
            The index to retrieve.

        Raises
        ------
        MissingDateError
            If the date is missing.

        Returns:
            NDArray[Any]: The data at the specified index.
        """
        raise MissingDateError(f"Date {self.dates[n]} is missing (index={n})")

    def tree(self) -> Node:
        """Builds a tree representation of the missing dataset.

        Returns
        -------
        Node
            The tree representation of the missing dataset.
        """
        return Node(self, [self.forward.tree()], start=self.start, end=self.end)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Provides metadata specific to the subclass.

        Returns
        -------
        Dict[str, Any]
            Metadata specific to the subclass.
        """
        return {"start": self.start, "end": self.end}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import xarray as xr
from anemoi.utils.testing import skip_if_offline
from anemoi.utils.testing import skip_slow_tests

from anemoi.datasets.create.sources.xarray import XarrayFieldList
from anemoi.datasets.testing import assert_field_list


@skip_if_offline
@skip_slow_tests
def test_opendap() -> None:
    """Test loading and validating the opendap dataset."""
    ds = xr.open_dataset(
        "https://thredds.met.no/thredds/dodsC/meps25epsarchive/2023/01/01/meps_det_2_5km_20230101T00Z.nc",
    )

    fs = XarrayFieldList.from_xarray(ds)
    assert_field_list(fs, 79529, "2023-01-01T00:00:00", "2023-01-03T18:00:00")


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Set
from typing import Union

import numpy as np
from anemoi.utils.dates import frequency_to_timedelta
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import Source
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import make_slice_or_index_from_list_or_tuple
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


def _default(a: int, b: int, dates: NDArray[np.datetime64]) -> list[int]:
    """Default combination function for reasons.

    Parameters:
    a (int): First integer value.
    b (int): Second integer value.
    dates (NDArray[np.datetime64]): Array of datetime64 dates.

    Returns:
    list[int]: List containing the two input integers.
    """
    return [a, b]


def _start(a: int, b: int, dates: NDArray[np.datetime64]) -> int:
    """Determine the start date between two dates.

    Parameters:
    a (int): First integer value.
    b (int): Second integer value.
    dates (NDArray[np.datetime64]): Array of datetime64 dates.

    Returns:
    int: The index of the start date.
    """
    from .misc import as_first_date

    c = as_first_date(a, dates)
    d = as_first_date(b, dates)
    if c < d:
        return b
    else:
        return a


def _end(a: int, b: int, dates: NDArray[np.datetime64]) -> int:
    """Determine the end date between two dates.

    Parameters:
    a (int): First integer value.
    b (int): Second integer value.
    dates (NDArray[np.datetime64]): Array of datetime64 dates.

    Returns:
    int: The index of the end date.
    """
    from .misc import as_last_date

    c = as_last_date(a, dates)
    d = as_last_date(b, dates)
    if c < d:
        return a
    else:
        return b


def _combine_reasons(reason1: Dict[str, Any], reason2: Dict[str, Any], dates: NDArray[np.datetime64]) -> Dict[str, Any]:
    """Combine two reason dictionaries.

    Parameters:
    reason1 (Dict[str, Any]): First reason dictionary.
    reason2 (Dict[str, Any]): Second reason dictionary.
    dates (NDArray[np.datetime64]): Array of datetime64 dates.

    Returns:
    Dict[str, Any]: Combined reason dictionary.
    """

    reason = reason1.copy()
    for k, v in reason2.items():
        if k not in reason:
            reason[k] = v
        else:
            func = globals().get(f"_{k}", _default)
            reason[k] = func(reason[k], v, dates)
    return reason


class Subset(Forwards):
    """Select a subset of the dates.

    Attributes:
    dataset (Dataset): The dataset.
    indices (List[int]): List of indices.
    reason (Dict[str, Any]): Dictionary of reasons.
    """

    def __init__(self, dataset: Union[Dataset, "Subset"], indices: Sequence[int], reason: Dict[str, Any]) -> None:
        """Initialize the Subset.

        Parameters:
        dataset (Dataset | Subset): The dataset or subset.
        indices (Sequence[int]): Sequence of indices.
        reason (Dict[str, Any]): Dictionary of reasons.
        """
        while isinstance(dataset, Subset):
            indices = [dataset.indices[i] for i in indices]
            reason = _combine_reasons(reason, dataset.reason, dataset.dates)
            dataset = dataset.dataset

        self.dataset: Dataset = dataset
        self.indices: List[int] = list(indices)
        self.reason: Dict[str, Any] = {k: v for k, v in reason.items() if v is not None}

        # Forward other properties to the super dataset
        super().__init__(dataset)

    def clone(self, dataset: Dataset) -> Dataset:
        """Clone the subset with a new dataset.

        Parameters:
        dataset (Dataset): The new dataset.

        Returns:
        Dataset: The cloned subset.
        """
        return self.__class__(dataset, self.indices, self.reason).mutate()

    def mutate(self) -> Dataset:
        """Mutate the subset.

        Returns:
        Dataset: The mutated subset.
        """
        return self.forward.swap_with_parent(parent=self)

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Get item by index.

        Parameters:
        n (FullIndex): The index.

        Returns:
        NDArray[Any]: The indexed data.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        assert n >= 0, n
        n = self.indices[n]
        return self.dataset[n]

    @debug_indexing
    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Get slice of data.

        Parameters:
        s (slice): The slice.

        Returns:
        NDArray[Any]: The sliced data.
        """
        # TODO: check if the indices can be simplified to a slice
        # the time checking maybe be longer than the time saved
        # using a slice
        indices = [self.indices[i] for i in range(*s.indices(self._len))]
        indices = make_slice_or_index_from_list_or_tuple(indices)
        if isinstance(indices, slice):
            return self.dataset[indices]
        return np.stack([self.dataset[i] for i in indices])

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, n: TupleIndex) -> NDArray[Any]:
        """Get tuple of data.

        Parameters:
        n (TupleIndex): The tuple index.

        Returns:
        NDArray[Any]: The indexed data.
        """
        index, changes = index_to_slices(n, self.shape)
        indices = [self.indices[i] for i in range(*index[0].indices(self._len))]
        indices = make_slice_or_index_from_list_or_tuple(indices)
        index, _ = update_tuple(index, 0, indices)
        result = self.dataset[index]
        result = apply_index_to_slices_changes(result, changes)
        return result

    def __len__(self) -> int:
        """Get the length of the subset.

        Returns:
        int: The length of the subset.
        """
        return len(self.indices)

    @cached_property
    def shape(self) -> Shape:
        """Get the shape of the subset."""
        return (len(self),) + self.dataset.shape[1:]

    @cached_property
    def dates(self) -> NDArray[np.datetime64]:
        """Get the dates of the subset."""
        return self.dataset.dates[self.indices]

    @cached_property
    def frequency(self) -> datetime.timedelta:
        """Get the frequency of the subset."""
        dates = self.dates
        if len(dates) < 2:
            raise ValueError(f"Cannot determine frequency of a subset with less than two dates ({self.dates}).")
        return frequency_to_timedelta(dates[1].astype(object) - dates[0].astype(object))

    def source(self, index: int) -> Source:
        """Get the source of the subset.

        Parameters:
        index (int): The index.

        Returns:
        Source: The source of the subset.
        """
        return Source(self, index, self.forward.source(index))

    def __repr__(self) -> str:
        """Get the string representation of the subset.

        Returns:
        str: The string representation of the subset.
        """
        return f"Subset({self.dataset},{self.dates[0]}...{self.dates[-1]}/{self.frequency})"

    @cached_property
    def missing(self) -> Set[int]:
        """Get the missing indices of the subset."""
        missing = self.dataset.missing
        result: Set[int] = set()
        for j, i in enumerate(self.indices):
            if i in missing:
                result.add(j)
        return result

    def tree(self) -> Node:
        """Get the tree representation of the subset.

        Returns:
        Node: The tree representation of the subset.
        """
        return Node(self, [self.dataset.tree()], **self.reason)

    def forwards_subclass_metadata_specific(self) -> Dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns:
        Dict[str, Any]: The metadata specific to the forwards subclass.
        """
        return {
            # "indices": self.indices,
            "reason": self.reason,
        }



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Hashable
from typing import List
from typing import Tuple

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, params: List[str], output: str) -> ekd.FieldList:
    """Computes the sum over a set of variables.

    Args:
        context (Any): The execution context.
        input (List[Any]): The list of input fields.
        params (List[str]): The list of parameters to sum over.
        output (str): The name for the output field.

    Returns:
        ekd.FieldList: The resulting FieldArray with summed fields.
    """
    result = []

    needed_fields: Dict[Tuple[Hashable, ...], Dict[str, ekd.Field]] = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        if param in params:
            key = tuple(key.items())

            if param in needed_fields[key]:
                raise ValueError(f"Duplicate field {param} for {key}")

            needed_fields[key][param] = f
        else:
            result.append(f)

    for keys, values in needed_fields.items():

        if len(values) != len(params):
            raise ValueError("Missing fields")

        s = None
        for k, v in values.items():
            c = v.to_numpy(flatten=True)
            if s is None:
                s = c
            else:
                s += c
        result.append(new_field_from_numpy(values[list(values.keys())[0]], s, param=output))

    return new_fieldlist_from_list(result)



# (C) Copyright 2025  Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any
from typing import Dict

import earthkit.data as ekd

from ..filter import Filter


class TransformFilter(Filter):
    """Calls filters from anemoi.transform.filters

    Parameters
    ----------
    context : Any
        The context in which the filter is created.
    name : str
        The name of the filter.
    config : Dict[str, Any]
        The configuration for the filter.
    """

    def __init__(self, context: Any, name: str, config: Dict[str, Any]) -> None:

        from anemoi.transform.filters import create_filter

        self.name = name
        self.transform_filter = create_filter(context, config)

    def execute(self, input: ekd.FieldList) -> ekd.FieldList:
        """Execute the transformation filter.

        Parameters
        ----------
        input : ekd.FieldList
            The input data to be transformed.

        Returns
        -------
        ekd.FieldList
            The transformed data.
        """
        return self.transform_filter.forward(input)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from earthkit.data import from_source

from anemoi.datasets.create.utils import to_datetime_list

from .legacy import legacy_source


@legacy_source(__file__)
def source(context: Optional[Any], dates: List[datetime], **kwargs: Any) -> Any:
    """Generates a source based on the provided context, dates, and additional keyword arguments.

    Parameters
    ----------
    context : Optional[Any]
        The context in which the source is generated.
    dates : List[datetime]
        A list of datetime objects representing the dates.
    **kwargs : Any
        Additional keyword arguments for the source generation.

    Returns
    -------
    Any
        The generated source.
    """
    name = kwargs.pop("name")
    context.trace("✅", f"from_source({name}, {dates}, {kwargs}")
    if kwargs["date"] == "$from_dates":
        kwargs["date"] = list({d.strftime("%Y%m%d") for d in dates})
    if kwargs["time"] == "$from_dates":
        kwargs["time"] = list({d.strftime("%H%M") for d in dates})
    return from_source(name, **kwargs)


execute = source

if __name__ == "__main__":
    import yaml

    config: Dict[str, Any] = yaml.safe_load(
        """
      name: mars
      class: ea
      expver: '0001'
      grid: 20.0/20.0
      levtype: sfc
      param: [2t]
      number: [0, 1]
      date: $from_dates
      time: $from_dates
    """
    )
    dates: List[str] = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    for f in source(None, dates, **config):
        print(f, f.to_numpy().mean())
